import os
import sys

from onepanel.utilities.s3_authenticator import S3Authenticator


class CredsUtility:
    @staticmethod
    def get_credentials(connection, account_uid, entity_type, entity_uid):
        """

        :param connection:
            - Usually from ctx.obj['connection']
        :param account_uid: string
        :param entity_type: string
        :param entity_uid: string
        :return:
        """
        cloud_provider = os.getenv('CLOUD_PROVIDER', 'AWS')
        if cloud_provider == "AWS":
            s3auth = S3Authenticator(connection)
            creds = s3auth.get_s3_credentials(account_uid, entity_type, entity_uid)
        else:
            creds = None
        if creds is None or creds['data'] is None:
            print("Unable to get provider credentials. Exiting.")
            sys.exit(-1)
        return creds


"""
Job commands
"""
import base64
import json
import os
import stat
import sys
import threading
import logging

import click
import configobj
import websocket

import onepanel.models.models as models
from onepanel.utilities.cloud_storage_utility import CloudStorageUtility
from onepanel.models.job import Job
from onepanel.models.metric import MetricSummary, Metric
from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.commands.projects import ProjectViewController
from onepanel.commands.datasets import DatasetViewController
from onepanel.onepanel_types.project_repository import PROJECT_REPOSITORY
from onepanel.onepanel_types.dataset_mount_identifier import DATASET_MOUNT_IDENTIFIER
from onepanel.utilities.s3.file_sync import FileSynchronizer, ThreadedFileSynchronizer, FileDifference

import onepanel.utilities.s3.wrapper as S3
from onepanel.utilities.s3.authentication import APIProvider, MemoryCachedCredentialsProvider

import onepanel.services as services


COMMON_LOGGER = 'common_logger'


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        if cmd_name == 'ls':
            return click.Group.get_command(self, ctx, 'list')

        return None


class JobViewController(APIViewController):
    JOB_OUTPUT_FILE = os.path.join('.onepanel','output')
    s3_bucket_name = 'onepanel-datasets'

    @staticmethod
    def get_cloud_provider_root_for_job_output(account_uid='', project_uid='', job_uid=''):
        return '{account_uid}/projects/{project_uid}/jobs/{job_uid}/output'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            job_uid=job_uid
        )

    @staticmethod
    def get_cloud_provider_compressed_file_for_job_output_path(account_uid='', project_uid='', job_uid=''):
        return '{account_uid}/projects/{project_uid}/jobs/{job_uid}/job-{job_uid}-output.tar.gz'.format(
            account_uid=account_uid,
            project_uid=project_uid,
            job_uid=job_uid
        )

    def __init__(self, conn):
        APIViewController.__init__(self, conn)
        self.job_uid = None
        self.project_account_uid = None
        self.project_uid = None

        self.initial_account_uid = conn.account_uid
        self.initial_project_uid = None

    def get_project_config(self):
        # Figure out the  account uid and project uid from file
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            return
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        if not os.path.exists(project_file):
            return

        cfg = configobj.ConfigObj(project_file)
        project_uid = cfg['uid']
        project_account_uid = cfg['account_uid']

        if len(project_uid) < 1 or len(project_account_uid) < 1:
            return

        self.project_account_uid = project_account_uid
        self.project_uid = project_uid
        self.initial_project_uid = project_uid

    def get_job_output_config(self):
        # Figure out the job uid, account uid and project uid from file
        home = os.getcwd()
        onepanel_dir = os.path.join(home, '.onepanel')
        if not os.path.exists(onepanel_dir):
            print("ERROR.Directory does not exist, cannot carry out all jobs operations.")
            print("DETAILS." + onepanel_dir)
            exit(-1)
        job_output_file = os.path.join(home, JobViewController.JOB_OUTPUT_FILE)
        if not os.path.isfile(job_output_file):
            print("ERROR.Job file does not exist, cannot carry out all jobs operations.")
            print("DETAILS." + job_output_file)
            exit(-1)

        cfg = configobj.ConfigObj(job_output_file)

        job_uid = cfg['uid']
        project_uid = cfg['project_uid']
        account_uid = cfg['account_uid']

        if len(job_uid) < 1 or len(project_uid) < 1 or len(account_uid) < 1:
            print("ERROR.Project file has invalid credentials. Verify credentials or re-pull project.")
            exit(-1)
        self.job_uid = job_uid
        self.project_uid = project_uid
        self.project_account_uid = account_uid

    def update_config(self, project_uid=None, account_uid=None, caller_is_sdk=False):
        if not project_uid and not self.initial_project_uid:
            print('Error: Project UID cannot be blank.')
            if caller_is_sdk:
                print('Set it with the `project_uid` parameter.')
                return False
            else:
                print('Set it with the --project option.')
                sys.exit(1)
        if not account_uid and not self.initial_account_uid:
            print('Error: Account UID cannot be blank.\nRun \'onepanel login\' to login.')
            if caller_is_sdk:
                return False
            else:
                sys.exit(1)

        self.project_uid = self.initial_project_uid
        self.project_account_uid = self.initial_account_uid
        if project_uid:
            self.project_uid = project_uid
        if account_uid:
            self.project_account_uid = account_uid

        self.endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/jobs'.format(
            root=self.conn.URL,
            account_uid=self.project_account_uid,
            project_uid=self.project_uid
        )

        return True

    def create(self, job):
        if job.code_repository is None:
            job.uid = self.upload_code()

        return self.post(post_object=job, json_encoder=models.APIJSONEncoder)

    def get_job(self, uid, project_uid=None, account_uid=None):
        response = self.get('/' + str(uid))
        if response['status_code'] == 404:
            print('Job not found')
            return None

        dataset_mounts_response = self.get('/' + str(uid) + '/dataset_mounts')
        if dataset_mounts_response['status_code'] == 200:
            response['data']['datasetMounts'] = dataset_mounts_response['data']

        return Job.from_json(response['data'])

    def get_job_metrics(self, uid):
        response = self.get('/' + str(uid) + '/metrics')

        if response['status_code'] in [401, 403]:
            raise APIViewController.UnauthorizedException('unauthorized')

        if response['status_code'] != 200:
            raise APIViewController.APIException(response['status_code'], "Unable to get job metrics")

        data = response['data']

        metric_summaries = [MetricSummary.from_json(datum) for datum in data]

        return metric_summaries

    def stop(self, account_uid=None, project_uid=None, job_uid=None):
        if account_uid is None and project_uid is None:
            account_uid = self.project_account_uid,
            project_uid = self.project_uid

        endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/job/'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            project_uid=project_uid
        )

        return self.delete_v2(job_uid, '/active', endpoint=endpoint)

    def delete_job(self, account_uid=None, project_uid=None, job_uid=None):
        if account_uid is None and project_uid is None:
            account_uid = self.project_account_uid,
            project_uid = self.project_uid

        endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/job/'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            project_uid=project_uid
        )

        return self.delete_v2(job_uid, endpoint=endpoint)

    def upload_code(self, uid=None, directory=os.getcwd()):
        endpoint = '{}/accounts/{}/projects/{}/credentials/aws'.format(self.conn.URL, self.project_account_uid, self.project_uid)
        authenticator = APIProvider(self.conn, endpoint)
        s3_wrapper = S3.Wrapper(credentials_provider=authenticator)

        print('Uploading files from current directory...')

        if not uid:
            res = self.put(endpoint='{}/accounts/{}/projects/{}/job_number'.format(self.conn.URL, self.project_account_uid, self.project_uid))
            if res['status_code'] != 200:
                print('There was an error uploading your files')
                return
            uid = str(res['data'])

        s3_wrapper.upload_directory(directory, '{}/projects/{}/jobs/{}/code/'.format(self.project_account_uid, self.project_uid, uid), exclude=['.git', '.onepanel'])
        print('Done uploading files.')
        
        return uid

    def download_code(self, uid, directory=os.getcwd()):
        endpoint = '{}/accounts/{}/projects/{}/credentials/aws'.format(self.conn.URL, self.project_account_uid, self.project_uid)
        authenticator = MemoryCachedCredentialsProvider(APIProvider(self.conn, endpoint))
        s3_wrapper = S3.Wrapper(credentials_provider=authenticator)

        print('Downloading code...')
        s3_wrapper.download_prefix('{}/projects/{}/jobs/{}/code/'.format(self.project_account_uid, self.project_uid, uid), directory)
        print('Code download complete')


@click.group(cls=AliasedGroup, help='Job commands')
@click.pass_context
def jobs(ctx):
    ctx.obj['vc'] = JobViewController(ctx.obj['connection'])
    ctx.obj['vc'].get_project_config()


@jobs.command('create', help='Create a job')
@click.argument(
    'command',
    type=str
)
@click.option(
    '-m', '--machine-type',
    type=str,
    required=True,
    help='Machine type UID. Run "onepanel machine-types list" for a list of UIDs.'
)
@click.option(
    '-e', '--environment',
    type=str,
    required=True,
    help='Environment UID. Run "onepanel environments list" for a list of UIDs.'
)
@click.option(
    '-s', '--storage',
    type=str,
    required=True,
    help='Volume type UID. Run "onepanel volume-types list" for a list of UIDs.'
)
@click.option(
    '-c', '--code-repository',
    type=PROJECT_REPOSITORY,
    required=False,
    help='''Repository and branch to pull code from.

            Format: url=<git-repo-url>[,branch=<branch>]

            Example: url=https://github.com/tensorflow/examples.git,branch=master

            "branch" is optional and will default to "master". 

            If omitted, code from current directory is uploaded.
         '''
)
@click.option(
    '-d', '--data',
    type=DATASET_MOUNT_IDENTIFIER,
    required=False,
    multiple=True,
    help='''Datasets to mount.

            Format: source=<account>/<dataset>[,version=<version>,destination=<mount-directory>]

            Example: source=onepanel/mnist,version=3,destination=mnist

            "version" is optional. If omitted, the latest version will be used.

            "destination" is optional. If omitted, "<account-name>/<dataset-name>/<version>" will be used.
        '''
)
@click.option(
    '--project',
    'project_uid',
    type=str,
    required=False,
    help='Project name'
)
@click.option(
    '--account',
    'account_uid',
    type=str,
    required=False,
    help='Account name'
)
@click.pass_context
@login_required
def create_job(ctx, command, machine_type, environment, storage, data, code_repository, project_uid=None, account_uid=None):
    vc = ctx.obj['vc']
    vc.update_config(project_uid, account_uid)

    # Get the latest versions for datasets where it is unspecified.
    dvc = DatasetViewController(ctx.obj['connection'])
    for datum in data:
        if datum.version is None:
            dvc.account_uid = vc.project_account_uid
            dvc.dataset_uid = datum.dataset_uid
            dvc.init_endpoint()
            datum.version = dvc.get_version()['data']['version']['version']

    job = Job()
    job.command = command
    job.machine_type.uid = machine_type
    job.instance_template.uid = environment
    job.volume_type.uid = storage
    job.dataset_mount_claims = data
    job.code_repository = code_repository

    response = vc.create(job)
    if response['status_code'] == 200:
        print('Job #{} created at https://{}/{}/projects/{}/jobs/{}'
            .format(response['data']['uid'], os.getenv('ONEPANEL_HOST', 'c.onepanel.io'), vc.project_account_uid, vc.project_uid, response['data']['uid']))
    else:
        print("An error occurred: {}".format(response['data']))


@jobs.command('get', help='Get job detail', hidden=True)
@click.argument(
    'job_uid',
    metavar='<job-uid>',
    type=str
)
@click.option(
    '--project',
    'project_uid',
    type=str,
    required=False,
    help='Project name'
)
@click.option(
    '--account',
    'account_uid',
    type=str,
    required=False,
    help='Account name'
)
@click.pass_context
@login_required
def get_job(ctx, job_uid, project_uid=None, account_uid=None):
    vc = ctx.obj['vc']
    vc.update_config(project_uid, account_uid)

    if not job_uid:
        print("Error: Job UID cannot be blank.")
        return

    job = vc.get_job(job_uid, project_uid, account_uid)
    print('UID: ' + job.uid + '\n'
        'Name: ' + job.name + '\n'
        'Command:\n' + job.command + '\n')


@jobs.command('list', help='Show a list of jobs')
@click.option(
    '-a', '--all',
    type=bool,
    is_flag=True,
    default=False,
    help='Include completed jobs'
)
@click.option(
    '--project',
    'project_uid',
    type=str,
    required=False,
    help='Project name'
)
@click.option(
    '--account',
    'account_uid',
    type=str,
    required=False,
    help='Account name'
)
@click.pass_context
@login_required
def list_jobs(ctx, all, project_uid=None, account_uid=None):
    vc = ctx.obj['vc']
    vc.update_config(project_uid, account_uid)

    items = vc.list(params='?running=true' if not all else '')
    if items == None or items['totalItems'] == 0:
        print('No jobs found. Use "--all" flag to retrieve completed jobs.')
        return

    items = [Job.from_json(item).simple_view() for item in items['data']]

    vc.print_items(items, fields=['uid', 'state', 'command'],
        field_names=['UID', 'STATE', 'COMMAND'])


@jobs.command('stop', help='Stop a job')
@click.argument(
    'job_uid',
    metavar='<job-uid>',
    type=str
)
@click.option(
    '--project',
    'project_uid',
    type=str,
    required=False,
    help='Project name'
)
@click.option(
    '--account',
    'account_uid',
    type=str,
    required=False,
    help='Account name'
)
@click.pass_context
@login_required
def stop_job(ctx, job_uid, project_uid=None, account_uid=None):
    vc = ctx.obj['vc']
    vc.update_config(project_uid, account_uid)
    ctx.obj['vc'].delete(job_uid, field_path='/active', message_on_success='Job stopped',
                         message_on_failure='Job not found')


@jobs.command('logs', help='Show job log')
@click.argument(
    'job_uid',
    metavar='<job-uid>',
    type=str
)
@click.option(
    '--project',
    'project_uid',
    type=str,
    required=False,
    help='Project name'
)
@click.option(
    '--account',
    'account_uid',
    type=str,
    required=False,
    help='Account name'
)
@click.pass_context
@login_required
def job_logs(ctx, job_uid, project_uid=None, account_uid=None):
    vc = ctx.obj['vc']
    vc.update_config(project_uid, account_uid)

    job_data = vc.get('/' + job_uid, field_path='/logs')
    if job_data['status_code'] != 400:
        if job_data['data'] is None:
            print('Job not found.')
            return
        log = job_data['data']
        print(log)
    else:
        # Streaming via WebSocket
        # See https://pypi.python.org/pypi/websocket-client/

        def on_message(ws, message):
            if message[0] == '0':
                message = base64.b64decode(message[1:]).decode('utf-8', 'ignore')
                sys.stdout.write(message)
                sys.stdout.flush()

        def on_error(ws, error):
            if isinstance(error, websocket.WebSocketConnectionClosedException):
                return

            if isinstance(error, KeyboardInterrupt):
                return

            if error.status_code == 502 or error.status_code == 503:
                print('Job {} is preparing'.format(job_uid))
            else:
                print(error)

        def on_close(ws):
            print('connection closed')

        def on_open(ws):
            def send_auth_token(*args):
                ws.send(json.dumps({'Authtoken': ''}))

            threading.Thread(target=send_auth_token).start()

        ws_url = '{ws_root}/{account_uid}/projects/{project_uid}/jobs/{job_uid}/logs/ws?id_token={token}'.format(
            ws_root='wss://c.onepanel.io',
            account_uid=vc.project_account_uid,
            project_uid=vc.project_uid,
            job_uid=job_uid,
            token=ctx.obj['connection'].token
        )

        ws = websocket.WebSocketApp(
            url=ws_url,
            on_message=on_message,
            on_open=on_open,
            on_error=on_error
        )

        ws.run_forever()

    return False


def jobs_download_output(ctx, path, directory, archive_flag):
    jvc = JobViewController(ctx.obj['connection'])
    #
    # Resource
    #
    dirs = path.split('/')

    # Job output: Method 2
    # <account_uid>/projects/<project_uid>/jobs/<job_uid>
    if len(dirs) == 5:
        try:
            account_uid, projects_dir, project_uid, jobs_dir, job_uid = dirs
            assert (projects_dir == 'projects') and (jobs_dir == 'jobs')
        except:
            print('Incorrect job path')
            return None
    else:
        print('Incorrect job uid')
        return None

    #
    # Directory
    #
    if directory is None or directory == '.':
        home = os.getcwd()
    else:
        home = os.path.join(os.getcwd(), directory)

    # Check how the job output is stored
    jvc.update_config(project_uid, account_uid)
    job_data = jvc.get('/' + job_uid)
    if job_data['data'] is None:
        print("Job not found.")
        return False

    cloud_provider_download_to_path = home
    util = CloudStorageUtility.get_utility(jvc.project_account_uid, 'projects', entity_uid=jvc.project_uid)

    if archive_flag is True:
        print("Attempting to download the compressed output file to {home} directory.".format(home=cloud_provider_download_to_path))

        cloud_provider_path_to_download_from = jvc.get_cloud_provider_compressed_file_for_job_output_path(account_uid, project_uid, job_uid)

        stats = util.get_prefix_stats(cloud_provider_path_to_download_from)

        if stats['files'] == 0:
            print("This job did not create any output or output was not saved. \n" +
                  "If you want to save and version control your output, modify your script to "
                  "save all output to the '/onepanel/output' directory.\n")
            return

        file_name = cloud_provider_path_to_download_from.split('/')[-1]
        util.download_file(cloud_provider_download_to_path + '/' + file_name, cloud_provider_path_to_download_from)
    else:
        print("Attempting to download output to {home} directory.".format(home=cloud_provider_download_to_path))
        cloud_provider_path_to_download_from = jvc.get_cloud_provider_root_for_job_output(
            account_uid, project_uid, job_uid)

        prefix = cloud_provider_path_to_download_from
        cloud_provider_path_to_download_from += '/'

        # Check if there any actual files to download from the output
        stats = util.get_prefix_stats(cloud_provider_path_to_download_from)
        if stats['files'] == 0:
            print("This job did not create any output or output was not saved. \n" +
                  "If you want to save and version control your output, modify your script to "
                  "save all output to the '/onepanel/output' directory.\n")
            return

        ignore_file_states = [FileDifference.State.DELETED]
        CloudStorageUtility.sync_files(jvc.project_account_uid, 'projects', jvc.project_uid,
                                       cloud_provider_download_to_path, cloud_provider_path_to_download_from,
                                       master=FileSynchronizer.REMOTE, s3_full_path=prefix,
                                       ignore_file_states=ignore_file_states)

    print("Finished downloading.")


@jobs.command('delete', help='Delete a job')
@click.argument(
    'job_uid',
    metavar='<job-uid>',
    type=str
)
@click.option(
    '--project',
    'project_uid',
    type=str,
    required=False,
    help='Project name'
)
@click.option(
    '--account',
    'account_uid',
    type=str,
    required=False,
    help='Account name'
)
@click.pass_context
@login_required
def delete_job(ctx, job_uid, project_uid=None, account_uid=None):
    vc = ctx.obj['vc']
    vc.update_config(project_uid, account_uid)
    ctx.obj['vc'].delete(job_uid, message_on_success='Job deleted', message_on_failure='Job not found')


def remove_readonly(func, path, excinfo):
    os.chmod(path, stat.S_IWRITE)
    func(path)


def upload_directory_raw(ctx, project_account_uid=None, project_uid=None, job_uid=None, source_directory=None, destination_directory=None,
                     delete=False):
    ctx.obj['vc'] = JobViewController(ctx.obj['connection'])
    vc = ctx.obj['vc']

    if project_account_uid is None or project_uid is None or job_uid is None:
        vc.get_job_output_config()
    else:
        vc.job_uid = job_uid
        vc.project_uid = project_uid
        vc.project_account_uid = project_account_uid

    if destination_directory is None:
        destination_directory = '{}/projects/{}/jobs/{}/output'.format(
            vc.project_account_uid, vc.project_uid, vc.job_uid)

    connection = services.get_connection()
    endpoint = '{}/accounts/{}/projects/{}/credentials/aws'.format(connection.URL, vc.project_account_uid,
                                                                   vc.project_uid)
    if 'authenticator' in ctx.obj:
        authenticator = ctx.obj['authenticator']
    else:
        authenticator = MemoryCachedCredentialsProvider(APIProvider(connection, endpoint))
        ctx.obj['authenticator'] = authenticator

    ignore_file_states = []
    if not delete:
        ignore_file_states.append(FileDifference.State.DELETED)

    s3_wrapper = S3.Wrapper(credentials_provider=authenticator)
    synchronizer = FileSynchronizer(source_directory, destination_directory, s3_wrapper, ignore_file_states=ignore_file_states)
    synchronizer = ThreadedFileSynchronizer(synchronizer)

    differences = synchronizer.find_difference()

    if len(differences) == 0:
        click.echo('No file differences')
        return

    synchronizer.synchronize(differences.values())
    synchronizer.shutdown()

    logging.getLogger(COMMON_LOGGER).info("Finished output upload.")


@jobs.command('upload-directory', help='Uploads directory content for a job', hidden=True)
@click.option('-a', '--project_account_uid', type=str, help='The uid of the account that owns the project the job is in')
@click.option('-p', '--project_uid', type=str, help='The uid of the project the job is for')
@click.option('-j', '--job_uid', type=str, help='The job uid')
@click.option('-s', '--path', type=str, help='Path to sync files from')
@click.option('-d', '--destination', type=str, help='Path to sync files to')
@click.option('-v', '--verbose', type=bool, help='If command should log what it is doing')
@click.option('-x', '--delete', type=bool, help='Should remote files delete to match local?')
@click.pass_context
@login_required
def upload_directory(ctx, project_account_uid, project_uid, job_uid, source_directory=None, destination_directory=None,
                     verbose=None, delete=False):
    if verbose is None:
        verbose = False

    if verbose:
        logger = logging.getLogger(COMMON_LOGGER)
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    upload_directory_raw(ctx, project_account_uid, project_uid, job_uid, source_directory, destination_directory, delete)


import json


class Metric:
    @classmethod
    def from_json(cls, dct):
        return cls(dct["x_axis"], dct["x_value"], dct["metric"], dct["value"])

    def __init__(self, x_axis, x_value, metric, value):
        """

        :param x_axis:
        :type x_axis str
        :param x_value:
        :type x_value float
        :param metric:
        :type metric str
        :param value:
        :type value float
        """
        self.x_axis = x_axis
        self.x_value = x_value
        self.metric = metric
        self.value = value


class MetricSummary:
    @classmethod
    def from_json(cls, dct):
        metrics = [Metric.from_json(raw_metric) for raw_metric in json.loads(dct["summaryData"])]
        summary = cls(dct["metric"], dct["displayOrder"], dct["xAxis"], dct["xStartValue"], dct["xFinalValue"],
                      dct["metricStartValue"], dct["metricFinalValue"], dct["dataUrl"], metrics)

        return summary

    def __init__(self, metric, display_order, x_axis, x_start_value,
                 x_final_value, metrics_start_value, metrics_final_value,
                 data_url, summary_data):
        """

        :param metric:
        :type metric str
        :param display_order:
        :type display_order int
        :param x_axis:
        :type x_axis str
        :param x_start_value:
        :type x_start_value float
        :param x_final_value:
        :type x_final_value float
        :param metrics_start_value:
        :type metrics_start_value float
        :param metrics_final_value:
        :type metrics_final_value float
        :param data_url:
        :type data_url str
        :param summary_data:
        :type summary_data list[Metric]
        """
        self.metric = metric
        self.display_order = display_order
        self.x_axis = x_axis
        self.x_start_value = x_start_value
        self.x_final_value = x_final_value
        self.metric_start_value = metrics_start_value
        self.metric_final_value = metrics_final_value
        self.data_url = data_url
        self.summary_data = summary_data




import os

from onepanel.utilities.cloud_storage_utility import CloudStorageUtility
from onepanel.utilities.s3.file_sync import FileSynchronizer, FileDifference
from onepanel.commands.jobs import JobViewController

from onepanel.models import Job


class Jobs:
    def __init__(self, conn):
        self.job_view_controller = JobViewController(conn)
        self.job_view_controller.project_account_uid = conn.account_uid

    def list(self, all=False, project_uid=None, account_uid=None):
        if not self.job_view_controller.update_config(project_uid, account_uid, caller_is_sdk=True):
            return

        jvc = self.job_view_controller

        items = jvc.list(params='?running=true' if not all else '')
        if items == None or items['totalItems'] == 0:
            msg = ['No jobs found.']
            if not all:
                msg.append(' Use "all=True" to retrieve completed jobs.')
            print(''.join(msg))
            return

        jobs = [Job.from_json(item).simple_view() for item in items['data']]
        jvc.print_items(jobs, fields=['uid', 'state', 'command'],
            field_names=['UID', 'STATE', 'COMMAND'])
        return [Job.from_json(item) for item in items['data']]

    def create(self, job):
        if not job:
            print("Error: Need a job object to create a job.")
            return
        if not job.command:
            print("Error: Job command must be provided.")
            return
        if not job.machine_type.uid:
            print("Error: Machine Type must be set.")
            return
        if not job.instance_template.uid:
            print("Error: Environment must be set.")
            return
        if not job.volume_type.uid:
            print("Error: A volume must be set.")
            return
        if not self.job_view_controller.update_config(job.project.uid, job.account.uid, caller_is_sdk=True):
            return

        response = self.job_view_controller.create(job)
        if response['status_code'] == 200:
            print('Created job: {uid}'.format(uid=response['data']['uid']))
            return response['data']['uid']
        else:
            print(response['data'])

    def stop(self, uid, project_uid=None, account_uid=None):
        if not uid:
            print("Error: Job UID cannot be blank.")
            return
        if not self.job_view_controller.update_config(project_uid, account_uid, caller_is_sdk=True):
            return

        response = self.job_view_controller.delete_v2('/' + str(uid) + '/active')
        if response['status_code'] == 200:
            print('Stopped job: {uid}'.format(uid=response['data']['uid']))
            return True
        else:
            print('Job is already stopped or does not exist.')
            return False

    def delete(self, uid, project_uid=None, account_uid=None):
        if not uid:
            print("Error: Job UID cannot be blank.")
            return
        if not self.job_view_controller.update_config(project_uid, account_uid, caller_is_sdk=True):
            return

        return self.job_view_controller.delete(uid, message_on_success='Deleted job', message_on_failure='Job not found')

    def get(self, uid, project_uid=None, account_uid=None):
        if not uid:
            print("Error: Job UID cannot be blank.")
            return
        if not self.job_view_controller.update_config(project_uid, account_uid, caller_is_sdk=True):
            return

        job = self.job_view_controller.get_job(uid, project_uid, account_uid)
        if job is not None:
            metrics = self.job_view_controller.get_job_metrics(uid)
            job.metrics = metrics

        return job

    def download_output(self, uid, archive_flag=False, project_uid=None, account_uid=None, target_directory=None):
        if not uid:
            print("Error: Job UID cannot be blank.")
            return
        if not self.job_view_controller.update_config(project_uid, account_uid, caller_is_sdk=True):
            return

        home = os.getcwd()
        jvc = self.job_view_controller
        cloud_provider_download_to_path = home
        if target_directory is not None:
            cloud_provider_download_to_path = target_directory

        util = CloudStorageUtility.get_utility(jvc.project_account_uid, 'projects', entity_uid=jvc.project_uid)
        if archive_flag is True:
            print('Attempting to download the compressed output file to {home} directory.'.format(
                home=cloud_provider_download_to_path))
            cloud_provider_path_to_download_from = jvc.get_cloud_provider_compressed_file_for_job_output_path(
                jvc.project_account_uid, jvc.project_uid, uid)

            stats = util.get_prefix_stats(cloud_provider_path_to_download_from)

            if stats['files'] == 0:
                print("This job did not create any output or output was not saved. \n" +
                      "If you want to save and version control your output, modify your script to "
                      "save all output to the '/onepanel/output' directory.\n")
                return

            file_name = cloud_provider_path_to_download_from.split('/')[-1]

            util.download_file(cloud_provider_download_to_path + '/' + file_name, cloud_provider_path_to_download_from)
        else:
            print('Attempting to download output to {home} directory.'.format(home=cloud_provider_download_to_path))
            cloud_provider_path_to_download_from = jvc.get_cloud_provider_root_for_job_output(
                jvc.project_account_uid, jvc.project_uid, uid)

            prefix = cloud_provider_path_to_download_from
            cloud_provider_path_to_download_from += '/'

            # Check if there any actual files to download from the output
            stats = util.get_prefix_stats(cloud_provider_path_to_download_from)
            if stats['files'] == 0:
                print("This job did not create any output or output was not saved. \n" +
                      "If you want to save and version control your output, modify your script to "
                      "save all output to the '/onepanel/output' directory.\n")
                return

            ignore_file_states = [FileDifference.State.DELETED]
            CloudStorageUtility.sync_files(jvc.project_account_uid, 'projects', jvc.project_uid,
                                           cloud_provider_download_to_path, cloud_provider_path_to_download_from,
                                           master=FileSynchronizer.REMOTE, s3_full_path=prefix,
                                           ignore_file_states=ignore_file_states)

        print('Finished downloading.')
        return True



import json


def login_helper(conn, email, username, password, token):
    if email != "" and username != "":
        print("You cannot use both email and username, pass in one or the other.")
        return
    if password != "" and token != "":
        print("You cannot use both password and token, pass in one or the other.")
        return
    if email == "" and username == "":
        print("An email or username must be passed in.")
        return
    if password == "" and token == "":
        print("A password or token must be passed in.")
        return

    url = conn.URL + '/sessions'
    data = {
        'sessions': [
            {'device': 'cli'}
        ],
        'account': {}
    }
    if email:
        data['email'] = email
    if username:
        data['uid'] = username
    if password:
        data['password'] = password
    if token:
        data['accessToken'] = token

    r = conn.put(url, data=json.dumps(data))
    if r.status_code == 200:
        data = r.json()
        return data
    elif r.status_code == 401 or r.status_code == 422:
        print('Incorrect credentials.')
        print('It may help to log out and then try to login.')
        return None
    else:
        print('Error: {}'.format(r.status_code))
        return None




import os
import sys
import json
import logging
from datetime import timedelta

import click


from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.commands.instances import InstanceViewController
from onepanel.commands.jobs import JobViewController

from onepanel.utilities.machine_activity import ThresholdMachineActivity
from onepanel.utilities.machine_activity import MachineActivityMonitor
from onepanel.utilities.machine_activity import MACHINE_ACTIVITY_MONITOR

import onepanel.utilities.s3.wrapper as S3
from onepanel.utilities.s3.authentication import APIProvider

class SystemViewController(APIViewController):
    def __init__(self, conn):
        APIViewController.__init__(self, conn)
        self.instance_view_controller = InstanceViewController(conn)
        self.job_view_controller = JobViewController(conn)

@click.group(help='System command', hidden=True)
@click.pass_context
def system(ctx):
    ctx.obj['vc'] = SystemViewController(ctx.obj['connection'])

@system.command(
    'monitor-activity',
    hidden=True,
    help='''Monitors the machine for activity. Needs 'account_uid/project_uid/workspace_uid' to identify machine. 
            If it is not active after some time, pause request is sent to API.'''
)
@click.argument(
    'identifier',
    type=str,
)
@click.option(
    '-c', '--cpu',
    type=float,
    help='Threshold for CPU. 40 = 40%. Anything below or equal to this is considered an inactive CPU'
)
@click.option(
    '-g', '--gpu',
    type=float,
    help='Threshold for GPU. 40 = 40%. Anything below or equal to this is considered an inactive GPU'
)
@click.option(
    '-d', '--duration',
    type=int,
    help='Duration, in seconds, after which request is sent to API to pause the machine, as it is inactive.'
)
@click.option(
    '-u', '--user-duration',
    type=int,
    help='Duration, in seconds, after which we consider user to be inactive. No commands within this, is inactive.'
)
@click.option(
    '-t', '--time-repeat',
    type=int,
    help='Duration, in seconds, in which the check is performed. E.g. every 5 seconds we check active status'
)
@click.option(
    '-v', '--verbose',
    type=bool,
    is_flag=True,
    default=False,
    help='If true, prints extra messages detailing status'
)
@click.pass_context
def monitor_activity(ctx, identifier, cpu=None, gpu=None, user_duration=None, duration=None, time_repeat=None, verbose=None):
    vc = ctx.obj['vc']

    ids = identifier.split('/')
    account_uid = ids[0]
    project_uid = ids[1]
    instance_uid = ids[2]

    if cpu is None:
        cpu = 40

    if gpu is None:
        gpu = 40

    if duration is None:
        duration = 3600

    if time_repeat is None:
        time_repeat = 60

    if user_duration is None:
        user_duration = 60 * 20  # 20 minutes

    logger = logging.getLogger(MACHINE_ACTIVITY_MONITOR)

    if verbose:
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    def pause_workspace():
        vc.instance_view_controller.pause(account_uid=account_uid, project_uid=project_uid, instance_uid=instance_uid)

    machine_activity = ThresholdMachineActivity(cpu, gpu, timedelta(seconds=user_duration),
                                                dead_duration=timedelta(seconds=duration))

    activity_monitor = MachineActivityMonitor(machine_activity, pause_workspace)
    activity_monitor.delay = time_repeat
    activity_monitor.start()
    activity_monitor.join()


@system.command(
    'print-activity',
    hidden=True,
    help='''Current machine status'''
)
@click.pass_context
def print_activity(ctx):
    machine_activity = ThresholdMachineActivity()
    click.echo(json.dumps(machine_activity.get_statistics(), indent=1))


@system.command('upload-job-code', hidden=True, help='Upload job code')
@click.argument('job_uid', metavar='<job-uid>', type=str)
@click.argument('directory', type=click.Path(), required=False)
@click.option(
    '--project',
    'project_uid',
    type=str,
    required=False,
    help='Project name'
)
@click.option(
    '--account',
    'account_uid',
    type=str,
    required=False,
    help='Account name'
)
@click.pass_context
@login_required
def upload_job_code(ctx, job_uid, directory=None, project_uid=None, account_uid=None):
    vc = ctx.obj['vc']
    vc.job_view_controller.update_config(project_uid, account_uid)

    if directory == None:
        directory = os.getcwd()
    vc.job_view_controller.upload_code(job_uid, directory)

@system.command('download-job-code', hidden=True, help='Download job code')
@click.argument('job_uid', metavar='<job-uid>', type=str)
@click.argument('directory', type=click.Path(), required=False)
@click.option(
    '--project',
    'project_uid',
    type=str,
    required=False,
    help='Project name'
)
@click.option(
    '--account',
    'account_uid',
    type=str,
    required=False,
    help='Account name'
)
@click.pass_context
@login_required
def download_job_code(ctx, job_uid, directory=None, project_uid=None, account_uid=None):
    vc = ctx.obj['vc']
    vc.job_view_controller.update_config(project_uid, account_uid)

    if directory == None:
        directory = os.getcwd()
    vc.job_view_controller.download_code(job_uid, directory)


import os
import threading
import time
import logging
import configobj

import click

from threading import Thread, Event

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from onepanel.utilities.cloud_storage_utility import CloudStorageUtility
from onepanel.utilities.timer import Timer
from onepanel.utilities.original_connection import Connection
from onepanel.utilities.dataset_api import DatasetAPI
from onepanel.utilities.s3.file_sync import FileSynchronizer


LOGGER = 'dataset-downloader'
# This is the directory datasets will be downloaded to if they are aliased.
DATASET_DOWNLOAD_DIRECTORY_IF_ALIAS = '.onepanel-datasets'


class DatasetDownloadStatus:
    """Watches a directory for files downloaded and keeps track of how many were downloaded
        and their statistics, like size.
    """
    def __init__(self, path):
        self.file_count = 0
        self.byte_count = 0
        self.path = path
        self.files_downloading = []

    def add_files(self, files):
        lock = threading.Lock()
        with lock:
            self.file_count += files

    def add_bytes(self, bytes_to_add):
        lock = threading.Lock()
        with lock:
            self.byte_count += bytes_to_add

    def set_files(self, value):
        lock = threading.Lock()
        with lock:
            self.file_count = value

    def set_bytes(self, value):
        lock = threading.Lock()
        with lock:
            self.byte_count = value

    def get_file_count(self):
        return self.file_count

    def get_byte_count(self):
        return self.byte_count

    def get_files_downloading(self):
        return self.files_downloading

    def add_file_downloading(self, file_path):
        lock = threading.Lock()
        with lock:
            self.files_downloading.append(file_path)

    def remove_file_downloading(self, file_path):
        lock = threading.Lock()
        with lock:
            self.files_downloading.remove(file_path)


class DatasetMountIdentifier:
    def __init__(self, account_uid, project_uid, resource_type, resource_uid):
        self.account_uid = account_uid
        self.project_uid = project_uid
        self.resource_type = resource_type
        self.resource_uid = resource_uid
        self.dataset_uid = ''
        self.dataset_version = 0
        self.dataset_mount_uuid = ''
        self.alias = None


class DatasetMountProgressUpdater:
    def __init__(self, api, stats, account_uid, project_uid, dataset_mount_uuid):
        self.stats = stats
        self.api = api
        self.account_uid = account_uid
        self.project_uid = project_uid
        self.dataset_mount_uuid = dataset_mount_uuid

    def update(self):
        total_bytes = self.calculate_total_bytes()
        total_files = self.stats.get_file_count()

        logging.getLogger(LOGGER).info('Sending Update to API files/bytes {}/{}'.format(total_files, total_bytes))
        response = self.api.update_dataset_mount_downloader(
            account_uid=self.account_uid,
            project_uid=self.project_uid,
            dataset_mount_uuid=self.dataset_mount_uuid,
            files_downloaded=total_files,
            bytes_downloaded=total_bytes)

        if response['status_code'] != 200:
            logging.getLogger(LOGGER).warning('Unable to update API. Response code is {}'.format(response['status_code']))

    def calculate_total_bytes(self):
        total_byte_count = self.stats.get_byte_count()

        for file_path in self.stats.get_files_downloading():
            try:
                total_byte_count += os.stat(file_path).st_size
            except:
                pass

        return total_byte_count


class DatasetDownloadStatusPrinter:
    def __init__(self, stats):
        self.stats = stats

    def print_stats(self):
        stats_string = 'Files: {files} Bytes: {bytes}'.format(files=self.stats.get_file_count(), bytes=self.stats.get_byte_count())
        click.echo(stats_string)


class AWSDownloadFileTracker(FileSystemEventHandler):
    def __init__(self, stats):
        self.stats = stats

    def is_temporary_file(self, file_path):
        # With AWS S3 CLI last dot has 8 hexadecimal characters after it.
        if file_path.count('.') == 0:
            return False

        last_dot_index = file_path.rfind('.')
        return len(file_path[last_dot_index + 1:]) == 8

    def on_created(self, event):
        if event.is_directory:
            return

        if self.is_temporary_file(event.src_path):
            self.stats.add_file_downloading(event.src_path)
            return

        # Otherwise, we have a full fledged downloaded file
        self.update_statistic(event.src_path)

    def on_moved(self, event):
        """For AWS, the file is downloaded as a temporary file if it's big.
           When it finishes, it renames it.
        """
        if event.is_directory:
            return

        self.stats.remove_file_downloading(event.src_path)
        self.update_statistic(event.dest_path)

    def update_statistic(self, file_path):
        logging.getLogger(LOGGER).info('Updating statistics with file {}'.format(file_path))

        self.stats.add_files(1)

        try:
            self.stats.add_bytes(os.stat(file_path).st_size)
        except Exception as argument:
            logging.getLogger(LOGGER).warning('Exception trying to update statistics {}'.format(argument))
            # Don't do anything
            pass


class DatasetDownloader(Thread):
    def __init__(self, owner_uid, dataset_mount, path, heartbeat=6.0, timer_delay=5.0):
        Thread.__init__(self)

        self.owner_uid = owner_uid
        self.path = path
        self.stop_flag = Event()
        self.heartbeat = heartbeat
        self.timer_delay = timer_delay
        self.timer = None
        self.dataset_mount = dataset_mount

        self.file_stats = None
        self.observer = None

        self.started = False

        self.api = None
        self.downloader_thread = None

    def start(self):
        Thread.start(self)

        conn = Connection()
        conn.load_credentials()

        self.api = DatasetAPI(conn)

        if not os.path.exists(self.path):
            os.makedirs(self.path)

        self.file_stats = DatasetDownloadStatus(self.path)
        self.observer = Observer()
        self.observer.schedule(AWSDownloadFileTracker(self.file_stats), self.path, True)

        self.downloader_thread = self.start_downloading_dataset(conn, self.path, self.dataset_mount.account_uid,
                                                                self.dataset_mount.dataset_uid,
                                                                self.dataset_mount.dataset_version)

        self.start_monitoring_files()
        self.start_updating_api(self.owner_uid, self.dataset_mount.project_uid,
                                self.dataset_mount.dataset_mount_uuid)

        self.started = True

    def run(self):
        while not self.stop_flag.wait(self.heartbeat):
            if self.started:
                self.downloader_thread.join()
                self.finished_downloading_hook()
                self.stop()

    def finished_downloading_hook(self):
        response = self.api.finish_dataset_mount_downloader(self.owner_uid, self.dataset_mount.project_uid,
                                                            self.dataset_mount.dataset_mount_uuid)

        response_code = response['status_code']
        if response_code != 200:
            logging.getLogger(LOGGER).warning('Error from finish mount: {}'.format(response_code))

        onepanel_dir = os.path.join(self.path, '.onepanel')
        if not os.path.exists(onepanel_dir):
            os.makedirs(onepanel_dir)

        dataset_file = os.path.join(self.path, os.path.join('.onepanel','dataset'))

        cfg = configobj.ConfigObj(dataset_file)
        cfg['uid'] = self.dataset_mount.dataset_uid
        cfg['account_uid'] = self.dataset_mount.account_uid
        cfg.write()

        if self.dataset_mount.alias is not None:
            symlink_path = self.dataset_mount.alias
            os.symlink(self.path, symlink_path, True)
            logging.getLogger(LOGGER).info('Created symlink {} -> {}'.format(self.path, symlink_path))

    def join(self, timeout=None):
        if self.downloader_thread is not None:
            self.downloader_thread.join()

        if self.timer is not None:
            self.timer.join()

        if self.observer is not None:
            try:
                self.observer.stop()
            except RuntimeError:
                print('observer was not started')

        Thread.join(self, timeout)

    def stop(self):
        if self.timer is not None:
            self.timer.stop()

        if self.observer is not None:
            try:
                self.observer.stop()
            except RuntimeError:
                print('observer was not started')

        self.stop_flag.set()

    def start_monitoring_files(self):
        self.observer.start()

    def start_updating_api(self, account_uid, project_uid, dataset_mount_uuid):
        updater = DatasetMountProgressUpdater(self.api, self.file_stats, account_uid, project_uid, dataset_mount_uuid)
        self.timer = Timer(self.timer_delay, updater.update)
        self.timer.start()

    def start_downloading_dataset(self, connection, download_to_path, account_uid, dataset_uid, dataset_version):
        current_version = self.api.get_dataset_version(account_uid, dataset_uid, dataset_version)

        if current_version['data'] is None:
            if current_version['status_code'] == 404:
                logging.getLogger(LOGGER).info('Dataset is without files.')
                return -1
            else:
                logging.getLogger(LOGGER).warning('Could not get information about dataset.')
                return -1

        if current_version['data']['provider']['uid'] != 'aws-s3':
            logging.getLogger(LOGGER).warning('Unsupported dataset provider')
            return -1

        cloud_storage_from_dir = current_version['data']['version']['path']
        remote_dataset_path = current_version['data']['version']['path'] + '/'
        t = Thread(target=CloudStorageUtility.sync_files, args=(account_uid, 'datasets', dataset_uid, download_to_path, remote_dataset_path,),
                                       kwargs={'master': FileSynchronizer.REMOTE, 'quiet': True, 's3_full_path': cloud_storage_from_dir, 'refresh': True})
        t.start()
        return t


class DatasetDownloadListener(Thread):
    """
        continue_listening: Whether the thread should continue waiting and check for new dataset mounts.
    """
    def __init__(self, connection, download_path, account_uid, project_uid,
                 resource_type, resource_uid, continue_listening=True, verbose=False, vc=None):
        Thread.__init__(self)

        self.identifier = DatasetMountIdentifier(account_uid, project_uid, resource_type, resource_uid)
        self.api = DatasetAPI(connection)
        self.download_path = download_path
        self.continue_listening = continue_listening
        self.vc = vc

        if verbose:
            logger = logging.getLogger(LOGGER)
            logger.setLevel(logging.DEBUG)
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            ch.setFormatter(formatter)
            logger.addHandler(ch)

    def get_dataset_mount(self):
        if self.identifier.resource_type == 'instance':
            return self.api.claim_instance_mount(
                        account_uid=self.identifier.account_uid,
                        project_uid=self.identifier.project_uid,
                        instance_uid=self.identifier.resource_uid,
                        downloader_id=os.getcwd(),
                        downloader_pid=os.getpid())
        elif self.identifier.resource_type == 'job':
            return self.api.claim_job_mount(
                        account_uid=self.identifier.account_uid,
                        project_uid=self.identifier.project_uid,
                        job_uid=self.identifier.resource_uid,
                        downloader_id=os.getcwd(),
                        downloader_pid=os.getpid())
        else:
            raise ValueError('Resource Type {} not supported'.format(self.identifier.resource_type))

    def run(self):
        logger = logging.getLogger(LOGGER)

        while True:
            logger.info('Listening for Dataset Mounts')

            try:
                response = self.get_dataset_mount()
            except Exception as Argument:
                # On exception, try again in 10 seconds
                logger.warning('Got an exception trying to claim an instance mount {}. Trying again in 10 seconds'.format(Argument))

                time.sleep(10)
                continue

            response_code = response['status_code']
            if response_code != 200:
                if response_code != 404:
                    logger.warning('Error from dataset mount: {}'.format(response_code))
                elif response_code == 404:
                    if not self.continue_listening:
                        logger.info('No more datasets to mount. Done.')
                        return
                    logger.info('No datasets to mount. Checking again in 30 seconds')
                # API says there is no more. Try back later.
                time.sleep(30)
                continue

            dataset_mount = response['data']

            download_identifier = DatasetMountIdentifier(dataset_mount['dataset']['account']['uid'],
                                                         self.identifier.project_uid,
                                                         self.identifier.resource_type,
                                                         self.identifier.resource_uid)

            download_identifier.dataset_mount_uuid = dataset_mount['uuid']
            download_identifier.dataset_uid = dataset_mount['dataset']['uid']
            download_identifier.dataset_version = dataset_mount['dataset_versioning']['version']

            path = os.path.join(self.download_path,
                                download_identifier.account_uid,
                                download_identifier.dataset_uid,
                                str(download_identifier.dataset_version))

            if dataset_mount['alias']:
                download_identifier.alias = os.path.join(self.download_path, dataset_mount['alias'])
                path = os.path.join(self.download_path, 
                                DATASET_DOWNLOAD_DIRECTORY_IF_ALIAS,
                                download_identifier.account_uid,
                                download_identifier.dataset_uid,
                                str(download_identifier.dataset_version))

            logger.info('Obtained dataset mount. Downloading to {}'.format(path))

            # Create the .onepanel folder
            self.vc.account_uid = download_identifier.account_uid
            self.vc.dataset_uid = download_identifier.dataset_uid
            self.vc.save(path)

            downloader = DatasetDownloader(self.identifier.account_uid, download_identifier, path)

            downloader.start()
            downloader.join()

            logger.info('Dataset {} Finished downloading'.format(path))




import sys

py3k = sys.version_info[0] >= 3

if py3k:
    import os

    def computer_cpu_count():
        return os.cpu_count()

    def make_dirs_if_not_exist(dirname):
        os.makedirs(dirname, exist_ok=True)

else:
    import os
    import multiprocessing

    def computer_cpu_count():
        return multiprocessing.cpu_count()

    def make_dirs_if_not_exist(dirname):
        try:
            os.makedirs(dirname)
        except os.error as exception:
            # exception is raised if some already exist, or can not be created.
            pass



from onepanel.models.api_json import APIJSON

class Environment(APIJSON):
    def __init__(self):
        self.uid = None
        self.name = None

    @classmethod
    def from_json(cls, dct):
        template = cls()

        template.uid = dct['uid']
        template.name = dct['name']

        return template

    def api_json(self):
        return {
            'uid': self.uid,
            'name': self.name
        }


import click
from onepanel.models.project_repository import ProjectRepository

class ProjectRepositoryParamType(click.ParamType):
    name = 'project_repository'

    def convert(self, value, param, ctx):
        """Accepts """
        try:
            project_repository = ProjectRepository()
            key_values = value.split(',')

            for kv in key_values:
                parts = kv.split('=')
                setattr(project_repository, parts[0], parts[1])
            
            return project_repository
        except ValueError:
            self.fail('%s is not valid' % value, param, ctx)

PROJECT_REPOSITORY = ProjectRepositoryParamType()


import os

from watchdog.events import PatternMatchingEventHandler

from onepanel.utilities.s3.file_sync import FileDifference, FileEvent


class FileWatchSynchronizerEventHandler(PatternMatchingEventHandler):
    def __init__(self, path, remote_path, synchronizer, ignore_patterns=None):
        """

        :param path:
        :param remote_path:
        :param synchronizer:
        :type synchronizer onepanel.utilities.s3.file_sync.ThreadedFileSynchronizer
        :param hook:
        """

        PatternMatchingEventHandler.__init__(self, ignore_patterns=ignore_patterns)

        self.path = path
        self.path_length = len(path) + 1
        self.remote_path = remote_path
        self.synchronizer = synchronizer

    def on_created(self, event):
        if event.is_directory:
            return

        relative_path = event.src_path[self.path_length:]

        if self.should_skip(relative_path):
            return

        destination_path = os.path.join(self.remote_path, relative_path)

        difference = FileDifference(event.src_path, destination_path, FileDifference.State.NEW)
        self._synchronize_difference(difference)

    def on_moved(self, event):
        if event.is_directory:
            return

        relative_path = event.src_path[self.path_length:]
        relative_destination = event.dest_path[self.path_length:]

        if self.should_skip(relative_path) or self.should_skip(relative_destination):
            return

        original_destination = os.path.join(self.remote_path, relative_path)
        destination = os.path.join(self.remote_path, relative_destination)

        difference = FileDifference(event.dest_path,
                                    destination,
                                    FileDifference.State.MOVED,
                                    event.src_path,
                                    original_destination)

        self._synchronize_difference(difference)

    def on_deleted(self, event):
        if event.is_directory:
            return

        relative_path = event.src_path[self.path_length:]

        if self.should_skip(relative_path):
            return

        destination_path = os.path.join(self.remote_path, relative_path)

        difference = FileDifference(event.src_path, destination_path, FileDifference.State.DELETED)
    
        self._synchronize_difference(difference)

    def on_modified(self, event):
        if event.is_directory:
            return

        relative_path = event.src_path[self.path_length:]

        if self.should_skip(relative_path):
            return

        destination_path = os.path.join(self.remote_path, relative_path)
        difference = FileDifference(event.src_path, destination_path, FileDifference.State.MODIFIED)
        self._synchronize_difference(difference)

    def _synchronize_difference(self, difference):
        self.synchronizer.synchronize_single(difference)

    def should_skip(self, relative_path):
        # Ignore vim temporary file and mac temporary files
        return relative_path[-1] == '~' or '.sb' in relative_path





from onepanel.models.api_json import APIJSON

class VolumeType(APIJSON):
    def __init__(self):
        self.uid = None
        self.info = {}

    def api_json(self):
        return {
            'uid': self.uid,
        }

    @classmethod
    def from_json(cls, dct):
        volume = cls()

        volume.uid = dct['uid']
        volume.info['size'] = dct['info']['size']

        return volume


import onepanel
from onepanel.utilities.login_helper import login_helper
from onepanel.utilities.original_connection import Connection


from onepanel.sdk.jobs import Jobs
from onepanel.sdk.machine_types import MachineTypes
from onepanel.sdk.volume_types import VolumeTypes
from onepanel.sdk.environments import Environments


class Client:
    def __init__(self, email="", username="", access_token=""):
        conn = Connection()
        user_id_present = False
        if email != "":
            user_id_present = True
        if username != "":
            user_id_present = True

        user_cred_present = False
        if access_token != "":
            user_cred_present = True

        if user_id_present and user_cred_present:
            data = login_helper(conn, email, username, "", access_token)
            if data is not None:
                conn.set_credentials(data['sessions'][0]['token'],data['account']['uid'])
            self.jobs = Jobs(conn)
            self.machine_types = MachineTypes(conn)
            self.volume_types = VolumeTypes(conn)
            self.environments = Environments(conn)
            return
        if user_id_present or user_cred_present:
            print("Not enough identifying arguments present.")
            return
        else:
            conn.load_credentials()
            self.jobs = Jobs(conn)
            self.machine_types = MachineTypes(conn)
            self.volume_types = VolumeTypes(conn)
            self.environments = Environments(conn)
            return



import datetime

from onepanel.models.api_json import APIJSON
from onepanel.utilities.time import parse_date, format_timedelta, utcnow
from onepanel.models.machine_type import MachineType
from onepanel.models.volume_type import VolumeType
from onepanel.models.instance_template import InstanceTemplate
from onepanel.models.dataset_mount import DatasetMount
from onepanel.models.project import Project
from onepanel.models.project_repository import ProjectRepository
from onepanel.models.account import Account


class Job(APIJSON):
    CREATED = "created"
    STARTING = "starting"
    READY = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    STOPPED = "stopped"

    @property
    def status(self):
        return self.state()

    def __init__(self):
        self.uid = None
        self.name = None
        self.created_at = None
        self.start_time = None
        self.completion_time = None
        self.active = 0
        self.succeeded = 0
        self.failed = 0
        self.log = None
        self.parent_uid = None
        self.user = None
        self.notes = None
        self.is_legacy = False
        self.dataset_mounts = []
        self.dataset_mount_claims = []
        self.uid = None
        self.command = None
        self.machine_type = MachineType()
        self.volume_type = VolumeType()
        self.instance_template = InstanceTemplate()
        self.project = Project()
        self.account = Account()
        self.code_repository = None
        self.environment = InstanceTemplate()
        if self.environment:
            self.instance_template = self.environment

        self.metrics = None

    def __str__(self):
        return str(self.__class__) + ": " + str(self.__dict__)

    @classmethod
    def from_json(cls, dct):
        job = cls()

        job.uid = dct['uid']
        job.name = dct['name']
        job.created_at = parse_date(dct['createdAt'])
        job.start_time = parse_date(dct['startTime'])
        job.completion_time = parse_date(dct['completionTime'])
        job.active = dct['active']
        job.succeeded = dct['succeeded']
        job.failed = dct['failed']
        job.log = dct['log']
        job.notes = dct['notes']
        job.is_legacy = dct['isLegacy']
        job.command = dct['command']
        job.machine_type = MachineType.from_json(dct['machineType'])
        job.volume_type = VolumeType.from_json(dct['volumeType'])
        job.instance_template = InstanceTemplate.from_json(dct['instanceTemplate'])
        if 'projectRepository' in dct:
            job.code_repository = ProjectRepository.from_json(dct['projectRepository'])
        if 'datasetMounts' in dct and dct['datasetMounts'] is not None:
            job.dataset_mount_claims = [DatasetMount.from_json(item) for item in dct['datasetMounts']]

        return job

    @classmethod
    def from_simple_json(cls, dct):
        job = cls()

        job.machine_type.uid = dct['machine']
        job.volume_type.uid = dct['storage']
        job.instance_template.uid = dct['environment']
        job.dataset_mount_claims = [DatasetMount.from_simple_json(item) for item in dct['datasets']]
        if 'projectRepository' in dct:
            job.code_repository = ProjectRepository.from_json(dct['projectRepository'])

        return job

#TODO document
    def simple_view(self):
        info = self.machine_type.info

        item = {
            'uid': self.uid,
            'name': self.name,
            'command': self.command,
            'cpu': info['cpu'],
            'gpu': info['gpu'],
            'ram': info['ram'],
            'hdd': self.volume_type.info['size'],
            'duration_running': format_timedelta(self.duration_running()),
            'age': format_timedelta(self.age()),
            'state': self.state()
        }

        return item

    def state(self):
        if self.failed != 0:
            return Job.FAILED

        if self.succeeded != 0 and self.completion_time is not None:
            return Job.COMPLETED

        if self.start_time is None and self.active == 0:
            return Job.CREATED

        if self.start_time is None and self.active != 0:
            return Job.STARTING

        if self.start_time is not None and self.active != 0:
            return Job.READY

        return Job.STOPPED

    def duration_running(self, when=None):
        """How long the instance has been running for. This is calculated from the time it was
           ready to when."""
        if when is None:
            when = utcnow()

        if self.start_time is None:
            return datetime.timedelta(0)

        return when - self.start_time

    def age(self, when=None):
        """How long the instance has been created to when"""
        if when is None:
            when = utcnow()

        if self.created_at is None:
            return datetime.timedelta(0)

        return when - self.created_at

    def api_json(self):
        json = {
            'command': self.command,
            'uid': self.uid,
            'name': self.name,
            'machineType': self.machine_type,
            'volumeType': self.volume_type,
            'instanceTemplate': self.instance_template,
            'datasetMountClaims': self.dataset_mount_claims,
            'projectRepository': self.code_repository,
            'datasetMounts': []
        }

        return json




import os

from onepanel.commands.base import APIViewController


class S3Authenticator(APIViewController):
	def __init__(self,conn):
		APIViewController.__init__(self, conn)

	'''
		entity_type is 'datasets' or 'projects'
		entity_uid is 'dataset_uid' or 'project_uid'
	'''
	def get_s3_credentials(self,account_uid='',entity_type='',entity_uid=''):
		if account_uid is '' or entity_type is '' or entity_uid is '':
			return {} # Empty JSON
		return self.get('','/accounts/{account_uid}/{entity_type}/{entity_uid}/credentials/aws'.format(
			account_uid=account_uid,
			entity_type=entity_type,
			entity_uid=entity_uid
		))



""" Command line interface for the OnePanel Machine Learning platform

Wrap git commands and provide transparent integration onepanel commands with the git.
"""

import glob
import os
import subprocess
import onepanel


class GitWrapper:
    def __init__(self):
        self.HOST = os.getenv('GITLAB_GIT_HOST', 'git.onepanel.io')

    def init(self,onepanel_user_uid,gitlab_token,home, account_uid, project_uid):
        git_cmd = 'git init'
        p = subprocess.Popen(git_cmd, cwd=home, shell=True)
        p.wait()

        git_cmd = 'git remote add onepanel https://{}:{}@{}/{}/{}.git'.format(
            onepanel_user_uid,
            gitlab_token,
            self.HOST,
            account_uid,
            project_uid
        )
        p = subprocess.Popen(git_cmd, cwd=home, shell=True)
        p.wait()

    def clone(self,onepanel_user_uid,gitlab_token, home, account_uid, project_uid, ext=''):
        git_cmd = 'git clone -o onepanel https://{}:{}@{}/{}/{}{}.git {}'.format(
            onepanel_user_uid,
            gitlab_token,
            self.HOST,
            account_uid,
            project_uid,
            ext,
            home
        )
        p = subprocess.Popen(git_cmd, shell=True)
        p.wait()

        return p.returncode

    def exclude(self, home, files):
        exclude_file = os.path.join(home, '.git/info/exclude')
        with open(exclude_file,'a+') as f:
            for file in files:
                f.write(file + '\n')


from onepanel.models.api_json import APIJSON

class Account(APIJSON):
    def __init__(self):
        self.uid = None

    def api_json(self):
        return {
            'uid': self.uid,
        }

    @classmethod
    def from_json(cls, dct):
        account = cls()

        account.uid = dct['uid']

        return account


import click
from onepanel.models.git_source import GitSource

class GitSourceParamType(click.ParamType):
    name = 'git source'

    def convert(self, value, param, ctx):
        """Accepts branch_name/commit=commit_hash"""
        try:
            return GitSource.from_string(value)
        except ValueError:
            self.fail('%s is not a valid git source' % value, param, ctx)

GIT_SOURCE = GitSourceParamType()


import click
from onepanel.models.dataset_mount import DatasetMount


class DatasetMountIdentifierParamType(click.ParamType):
    name = 'dataset mount'

    def convert(self, value, param, ctx):
        """Accepts multiple key/value pairs separated by commas. The key/value is separated by equals sign.
           Order of the parameters doesn't matter.
           
           For example: source=andreyonepanel/images,version=3,destination=cats

           Source is required.
           Destination is optional. Defaults to None.
           Version is optional. Defaults to None, which indicates "use the latest".
        """
        try:
            identifier = DatasetMount()
            key_values = value.split(',')
            map = {}

            for key_value in key_values:
                parts = key_value.split('=')
                map[parts[0]] = parts[1]

            if 'version' not in map:
                map['version'] = None

            identifier.set_source(map['source'])
            identifier.set_version(map['version'])
            if 'destination' in map:
                identifier.destination = map['destination']

            return identifier
        except ValueError:
            self.fail('%s is not a valid dataset mount' % value, param, ctx)


DATASET_MOUNT_IDENTIFIER = DatasetMountIdentifierParamType()



from onepanel.models.api_json import APIJSON

class GitSource(APIJSON):
    def __init__(self, branch, commit_hash=None):
        self.version = 1
        self.branch = branch
        self.commit_hash = commit_hash

    @classmethod
    def from_json(cls, dct):
        data = dct['data']

        branch_name = data['branch']['name']

        if isinstance(data['commit'], dict):
            commit_id = data['commit']['id']
        else:
            commit_id = ''

        return cls(branch_name, commit_id)

    @classmethod
    def from_string(cls, value):
        """value is expected to be branch_name/commit_hash"""

        parts = value.split('/')

        return cls(parts[0], parts[1])

    def api_json(self):
        return {
            'version': self.version,
            'data': {
                'commit': {
                    'id': self.commit_hash
                },
                'branch': {
                    'name': self.branch
                }
            }
        }


"""
Workspace module
"""
import os
import sys
import json
import logging

import click
import configobj

import onepanel.models.models as models
from onepanel.models.instance import Instance
from onepanel.commands.base import APIViewController
from onepanel.commands.login import login_required
from onepanel.commands.projects import ProjectViewController
from onepanel.onepanel_types.duration import DURATION
from onepanel.onepanel_types.dataset_mount_identifier import DATASET_MOUNT_IDENTIFIER
from onepanel.onepanel_types.git_source import GIT_SOURCE

from datetime import timedelta


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        if cmd_name == 'ls':
            return click.Group.get_command(self, ctx, 'list')

        return None


class InstanceViewController(APIViewController):
    @staticmethod
    def get_credentials_from_current_path():
        home = os.getcwd()
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        cfg = configobj.ConfigObj(project_file)

        if len(cfg) == 0:
            return None

        return {
            'account_uid': cfg['account_uid'],
            'project_uid': cfg['uid']
        }

    def __init__(self, conn):
        APIViewController.__init__(self, conn)
        self.account_uid = None
        self.uid = None

    def init_credentials_retrieval(self):
        home = os.getcwd()
        project_file = os.path.join(home, ProjectViewController.PROJECT_FILE)
        cfg = configobj.ConfigObj(project_file)

        if len(cfg) == 0:
            click.echo("ERROR.Cannot figure out the current project.")
            click.echo("Make sure .onepanel/project exists and is accurate.")
            exit(-1)
        self.account_uid = cfg['account_uid']
        self.uid = cfg['uid']

    def init_endpoint(self):
        self.endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=self.account_uid,
            project_uid=self.uid
        )

    def list(self, account_uid=None, project_uid=None, show_all=False):
        """List all of the workspaces for the provided account_uid, project_uid.
           If show_all is true, and there is no project_uid or account_uid, it will include all workspaces
           for projects the user is a member of too.
        """
        params = None

        if project_uid is not None and account_uid is not None:
            endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
                root=self.conn.URL,
                account_uid=account_uid,
                project_uid=project_uid
            )
        else:
            endpoint = '{root}/instances'.format(
                root=self.conn.URL,
                account_uid=account_uid
            )

        if show_all:
            params = '?show_all=true'

        items = APIViewController.list(self, endpoint=endpoint, params=params)

        if items is None:
            return items

        return [Instance.from_json(item) for item in items]

    def create(self, instance):
        return self.post(post_object=instance, json_encoder=models.APIJSONEncoder)

    def pause(self, account_uid=None, project_uid=None, instance_uid=None):
        if account_uid is None and project_uid is None:
            account_uid = self.account_uid
            project_uid = self.uid

        endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            project_uid=project_uid
        )

        return self.get('/' + instance_uid, '/stop', endpoint=endpoint)

    def resume(self, account_uid=None, project_uid=None, instance_uid=None):
        if account_uid is None and project_uid is None:
            account_uid = self.account_uid,
            project_uid = self.uid

        endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            project_uid=project_uid
        )

        return self.get('/' + instance_uid, '/resume', endpoint=endpoint)

    def delete_instance(self, account_uid=None, project_uid=None, instance_uid=None):
        if account_uid is None and project_uid is None:
            account_uid = self.account_uid
            project_uid = self.uid

        endpoint = '{root}/accounts/{account_uid}/projects/{project_uid}/instances'.format(
            root=self.conn.URL,
            account_uid=account_uid,
            project_uid=project_uid
        )

        return self.delete_v2(instance_uid, endpoint=endpoint)


@click.group(cls=AliasedGroup, help='Workspace commands', hidden=True)
@click.pass_context
def workspaces(ctx):
    ctx.obj['vc'] = InstanceViewController(ctx.obj['connection'])


@workspaces.command(
    'create',
    help='Create a new workspace. The workspace\'s name: Max 25 chars, lower case alphanumeric or "-", '
         'must start and end with alphanumeric'
)
@click.argument(
    'workspace_uid',
    type=str
)
@click.option(
    '-m', '--machine-type',
    type=str,
    required=True,
    help='Machine type UID. Call "onepanel machine-types list" for UIDs.'
)
@click.option(
    '-e', '--environment',
    type=str,
    required=True,
    help='Instance template UID. Call "onepanel environments list" for UIDs.'
)
@click.option(
    '-s', '--storage',
    type=str,
    required=True,
    help='Storage type UID.'
)
@click.option(
    '-g', '--source_control',
    type=GIT_SOURCE,
    required=False,
    help='''Source control to use. 

            Format: branch_name/commit_hash

            If omitted, latest commit from master is used.
         '''
)
@click.option(
    '-d', '--mount',
    type=DATASET_MOUNT_IDENTIFIER,
    required=False,
    multiple=True,
    help='''Datasets to mount. 
    
            Format: source=account_name/dataset_name,version=version_of_dataset,destination=folder_to_call_it

            Example: source=onepaneldemo/images,version=3,destination=cats

            The version is optional. If omitted, it'll use the latest version.
            
            The destination is optional. If omitted, it'll use account_name/dataset_name/version
        '''
)
@click.pass_context
@login_required
def create_instance(ctx, workspace_uid, machine_type, environment, storage, mount, source_control):
    try:
        instance = Instance()
        instance.set_uid(workspace_uid)
        instance.machine_type.uid = machine_type
        instance.instance_template.uid = environment
        instance.volume_type.uid = storage
        instance.dataset_mount_claims = mount
        instance.gitSource = source_control

        response = create_instance_internal(ctx.obj['vc'], instance)
        if response['status_code'] == 200:
            click.echo('New workspace created: {}'.format(response['data']['uid']))
        else:
            click.echo('An error occurred with creating a new workspace.')
            click.echo("Details: {}".format(response['data']))

    except ValueError as error:
        click.echo('\tFailed - {}'.format(error))


@workspaces.command(
    'create-from-file',
    help='''Create new workspaces from a json file. 
            The JSON file contains an array. You can have muliple workspaces defined.
        
            \b
            sample.json
            [
                {
                    "name": "test-5",
                    "environment": "jupyter-py3-tensorflow1.11.0",
                    "machine" : "aws-shared",
                    "storage" : "default-storage-10",
                    "git": "master/9ee6aed97073a543b8f8cb56bbab613716f093a1",
                    "datasets": [
                        {
                            "source": "vafilor2/cats",
                            "version": 1,
                            "destination": "cats"
                        }
                    ]  
                }
            ]
            
            Each workspace is created sequentially, top down, one after another. If any fails, they are skipped. 
        '''
)
@click.argument(
    'file_path',
    type=click.Path(exists=True, dir_okay=False)
)
@click.pass_context
@login_required
def create_instances_from_file(ctx, file_path):
    with open(file_path, newline='') as file:
        instances = []

        try:
            instances = json.load(file)
        except json.decoder.JSONDecodeError as error:
            click.echo('Unable to parse json. {}'.format(error))
            return

        for instance in instances:
            click.echo('Creating instance {}'.format(instance['name']))
            try:
                model = Instance.from_simple_json(instance)
                response = create_instance_internal(ctx.obj['vc'], model)
                if response['status_code'] == 200:
                    click.echo('\tSuccess')
                else:
                    click.echo('\tFailed - {}'.format(response['data']))
            except ValueError as error:
                click.echo('\tFailed - {}'.format(error))


def create_instance_internal(vc, instance):
    vc.init_credentials_retrieval()
    vc.init_endpoint()

    return vc.create(instance)


@workspaces.command(
    'list',
    help='''Show workspaces. If this is done outside a project, it will show 
            all of the user's workspaces. If the --all flag is supplied, it will also show workspaces
            for projects the user is a member of.

            If this is done inside a project, only workspaces for that project are listed.
         '''
                    )
@click.option(
    '-a',
    '--show_all',
    is_flag=True,
    default=False,
    help='''If true, shows workspaces for projects the user is a member of.
            This has no effect if you are listing workspaces inside a project.
         '''
              )
@click.option(
    '-q',
    '--id_only',
    is_flag=True,
    default=False,
    help='''Only prints the id information. Inside a project, this is just the workspace uid.
            Outside the project this is a tuple [account_uid, project_uid, workspace_uid]

            No message is printed if no workspaces are found.
         '''
              )
@click.option(
    '-g',
    '--duration_gt',
    type=DURATION,
    help='''Only prints the id information. Inside a project, this is just the workspace uid.
            Outside the project this is account_uid/project_uid/workspace_uid

            No message is printed if no workspaces are found.
         '''
              )
@click.pass_context
@login_required
def list_instances(ctx, show_all, id_only, duration_gt):
    vc = ctx.obj['vc']

    account_uid = None
    project_uid = None

    project_credentials = vc.get_credentials_from_current_path()
    if project_credentials is not None:
        account_uid = project_credentials['account_uid']
        project_uid = project_credentials['project_uid']

    list_instances_internal(vc, account_uid, project_uid, show_all, id_only, duration_gt)


def list_instances_internal(vc, account_uid, project_uid, show_all, id_only, duration_gt=None):
    items = vc.list(account_uid, project_uid, show_all)

    if items is None or len(items) == 0:
        if not id_only:
            click.echo('No workspaces found.')

        return

    if duration_gt is not None:
        if sys.version_info[0] < 3:
            items = filter(lambda item: item.age() > duration_gt, items)
        else:
            items = list(filter(lambda item: item.age() > duration_gt, items))

    formatted_items = [item.simple_view() for item in items]

    project_specific = account_uid is not None and project_uid is not None

    if id_only and not project_specific:
        click.echo(" ".join('{}/{}/{}'.format(item.project.account.uid, item.project.uid, item.uid) for item in items))
    elif id_only and project_specific:
        click.echo(" ".join(item.uid for item in items))
    elif project_specific:
        vc.print_items(formatted_items,
                       fields=['uid', 'state', 'age', 'duration_ready', 'duration_paused', 'cpu', 'gpu', 'ram',
                               'hdd', ],
                       field_names=['UID', 'STATE', 'AGE', 'RUNNING', 'PAUSED', 'CPU', 'GPU', 'RAM', 'HDD'])
    else:
        vc.print_items(formatted_items,
                       fields=['account_uid', 'project_uid', 'uid', 'state', 'age', 'duration_ready', 'duration_paused',
                               'cpu', 'gpu', 'ram', 'hdd', ],
                       field_names=['ACCOUNT', 'PROJECT', 'UID', 'STATE', 'AGE', 'RUNNING', 'PAUSED', 'CPU', 'GPU',
                                    'RAM', 'HDD'])


@workspaces.command(
    'terminate',
    help='''Terminate the workspace(s) given the uids of the workspace
    
            If executing this command inside a project, just the workspace uid is needed.
            Otherwise, you need account_uid/project_uid/workspace_uid
         '''
)
@click.argument(
    'uids',
    type=str,
    nargs=-1
)
@click.pass_context
@login_required
def terminate_instance(ctx, uids):
    vc = ctx.obj['vc']

    if len(uids) == 0:
        click.echo('No uids passed in. No instances deleted')
        return

    for uid in uids:
        response = None

        if vc.get_credentials_from_current_path() is not None:
            vc.init_credentials_retrieval()
            vc.init_endpoint()
            response = vc.delete_instance(instance_uid=uid)
        else:
            ids = uid.split('/')
            response = vc.delete_instance(ids[0], ids[1], ids[2])

        if response['status_code'] == 200:
            click.echo('workspace {} deleted'.format(uid))
        elif response['status_code'] == 404:
            click.echo('workspace {} not found'.format(uid))
        else:
            click.echo('An error occurred with deleting workspace {}.'.format(uid))
            click.echo("Details: {}".format(response['data']))


@workspaces.command(
    'pause',
    help='''Pause the workspace(s) given the uids of the workspace
    
            If executing this command inside a project, just the workspace uid is needed.
            Otherwise, you need account_uid/project_uid/workspace_uid
         '''
)
@click.argument(
    'uids',
    type=str,
    nargs=-1
)
@click.pass_context
@login_required
def pause_instance(ctx, uids):
    vc = ctx.obj['vc']

    if len(uids) == 0:
        click.echo('No uids passed in. No instances paused')
        return

    for uid in uids:
        response = None

        if vc.get_credentials_from_current_path() is not None:
            vc.init_credentials_retrieval()
            vc.init_endpoint()
            response = vc.pause(instance_uid=uid)
        else:
            ids = uid.split('/')
            response = vc.pause(ids[0], ids[1], ids[2])

        if response['status_code'] == 200:
            click.echo('workspace {} paused'.format(uid))
        elif response['status_code'] == 404:
            click.echo('workspace {} not found'.format(uid))
        else:
            click.echo('An error occurred with pausing workspace {}.'.format(uid))
            click.echo("Details: {}".format(response['data']))


@workspaces.command(
    'resume',
    help='''Resume the workspace(s) given the uids of the workspace
    
            If executing this command inside a project, just the workspace uid is needed.
            Otherwise, you need account_uid/project_uid/workspace_uid
         '''
                    )
@click.argument(
    'uids',
    type=str,
    nargs=-1
)
@click.pass_context
@login_required
def resume_instance(ctx, uids):
    vc = ctx.obj['vc']

    if len(uids) == 0:
        click.echo('No uids passed in. No instances resume')
        return

    for uid in uids:
        response = None

        if vc.get_credentials_from_current_path() is not None:
            vc.init_credentials_retrieval()
            vc.init_endpoint()
            response = vc.pause(instance_uid=uid)
        else:
            ids = uid.split('/')
            response = vc.pause(ids[0], ids[1], ids[2])

        if response['status_code'] == 200:
            click.echo('workspace {} resumed'.format(uid))
        elif response['status_code'] == 404:
            click.echo('workspace {} not found'.format(uid))
        else:
            click.echo('An error occurred with resuming workspace {}.'.format(uid))
            click.echo("Details: {}".format(response['data']))




import sys
import os
import subprocess

CREATE_NO_WINDOW = 0x08000000
CREATE_NEW_PROCESS_GROUP = 0x00000200


def run_in_background(cmd_list):
    """Run a process in the background, cmd_list is the list of commands to execute"""
    close_fds = False

    if sys.platform != 'win32':
        cmd_list.insert(0, 'nice')
        cmd_list.insert(0, 'nohup')
        close_fds = True
    else:
        # /i so that windows doesn't create "%SYSTEM_DRIVE%" folder
        cmd_list.insert(0, 'start /b /i')

    cmd = ' '.join(cmd_list)
    if sys.platform != 'win32':
        stdout = open(os.path.devnull, 'a')
        stderr = open(os.path.devnull, 'a')
        subprocess.Popen(args=cmd, stdin=subprocess.PIPE, stdout=stdout,
                         stderr=stderr, shell=True, close_fds=close_fds, preexec_fn=os.setpgrp)
    else:
        # Windows specific instructions
        # https://docs.microsoft.com/en-us/windows/desktop/ProcThread/process-creation-flags
        subprocess.Popen(args=cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, shell=True, close_fds=close_fds,
                         creationflags=CREATE_NO_WINDOW | CREATE_NEW_PROCESS_GROUP)



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import concurrent.futures
import os.path
import datetime
import sys
import re

from onepanel.utilities.file import get_file_tree
from onepanel.utilities.time import UTC

from os.path import join
from threading import Lock
from onepanel.utilities.python_bridge import computer_cpu_count


class FileDifference:
    class State:
        MODIFIED = 'modified'
        NEW = 'new'
        DELETED = 'deleted'
        MOVED = 'moved'
    
    def __init__(self, source_path, destination_path, state, original_source_path=None, original_destination_path=None):
        """
        :param source_path: The source path of the file, assumed to be a filesystem path
        :type source_path: str
        :param destination_path: The destination path of the file, format is dependent on service used, like S3.
        :type destination_path: str
        :param state: The state of the file difference, NEW|MODIFIED|DELETED|MOVED. We assume that the difference
                      is from source to destination. E.g. source file is NEW relative to destination.
        :type state: str (one of the constants in FileDifference.State)
        :param original_source_path original path of file, if renaming or moving file.
        :type original_source_path str
        """

        if original_source_path is None:
            original_source_path = source_path

        if original_destination_path is None:
            original_destination_path = destination_path

        self.source_path = source_path
        self.original_source_path = original_source_path
        self.original_destination_path = original_destination_path
        self.destination_path = destination_path
        self.state = state

    def __str__(self):
        return "{{original_source: {}\nsource: {}\noriginal_destination: {}\ndestination: {}\nstate:{}}}"\
            .format(self.original_source_path, self.source_path,
                    self.original_destination_path, self.destination_path, self.state)


class FileEvent:
    START = 'start'
    FINISHED = 'finished'
    FAILED = 'failed'

    def __init__(self, state, file_difference, result=None):
        self.state = state
        self.file_difference = file_difference
        self.result = result


class FileSynchronizer:
    LOCAL = 0
    REMOTE = 1

    @staticmethod
    def filter_path(file_differences, path, prefix=None):
        if path[0] == '/':
            path = path[1:]

        result = {}
        for key, file_difference in file_differences.items():
            name = file_difference.destination_path
            if prefix is not None and prefix in name:
                name = name[len(prefix):]

            if name.startswith(path):
                result[file_difference.destination_path] = file_difference

        return result

    @staticmethod
    def filter(file_differences, regex, prefix=None):
        """
        :param file_differences: a map where name is a
        :param regex:
        :param prefix:
        :return:
        """

        result = {}
        pattern = re.compile(regex)
        for key, file_difference in file_differences.items():
            name = file_difference.destination_path
            if prefix is not None and prefix in name:
                name = name[len(prefix):]

            if pattern.search(name) is not None:
                result[name] = file_difference

        return result

    @staticmethod
    def local_file_stats(filepath):
        return {
            'last_modified': datetime.datetime.fromtimestamp(os.path.getmtime(filepath), UTC),
            'size': os.path.getsize(filepath)
        }
    
    @staticmethod
    def s3_file_stats(api_content):
        return {
            'last_modified': api_content['LastModified'],
            'size': api_content['Size']
        }

    @staticmethod
    def content_modification_difference_local(a, b):
        """
        Files are different if a has been modified after b, or, if equal,
        if the file sizes are different.
        :param a: local file
        :param b: remote file
        :return:
        """
        if a['last_modified'] > b['last_modified']:
            return True

        return a['size'] != b['size']

    @staticmethod
    def content_modification_difference_remote(a, b):
        """
        Files are different if a has been modified after b, or, if equal,
        if the file sizes are different.
        :param a: local file
        :param b: remote file
        :return:
        """
        return a['size'] != b['size']

    def __init__(self, filepath, s3_prefix, s3_wrapper, master=LOCAL, s3_full_path=None, ignore_file_states=None):
        """
        :param filepath:
        :type filepath str
        :param s3_prefix:
        :type s3_prefix str
        :param s3_wrapper:
        :type s3_wrapper onepanel.utilities.s3.wrapper.Wrapper
        :param master: Determines if the local files or the remote files are considered to be
                       the "master" and the opposite side should change its files to match it.
        :type master int one of the constants in FileSynchronizer
        :param s3_full_path: The full s3 path to get files from. Used to select subfolders.
                In this case s3_prefix is the "root" of the files, so the names will be relative to that.
        :param ignore_file_states: A list of file states to ignore. These will not be synchronized, etc.
        :param include_glob: A glob pattern of files to include. Files not matching this pattern will not be included.
        :type str
        """
        self.filepath = filepath
        self.s3_prefix = s3_prefix
        self.master = master
        self.s3_wrapper = s3_wrapper
        self.s3_full_path = s3_full_path

        if ignore_file_states is None:
            ignore_file_states = []

        self.ignore_file_states = ignore_file_states
        self.running = True

    def find_difference(self, comparator=None):
        """ Finds the differences in files between the file_path and s3_prefix using
        the provided master.
        :return: a map of the file differences. Key is file path locally, value is a FileDifference
        :type {}
        """

        if comparator is None and self.master == FileSynchronizer.LOCAL:
            comparator = FileSynchronizer.content_modification_difference_local
        elif comparator is None and self.master == FileSynchronizer.REMOTE:
            comparator = FileSynchronizer.content_modification_difference_remote

        if self.master == FileSynchronizer.LOCAL:
            return self._find_difference_local(comparator)
        else:
            return self._find_difference_remote(comparator)

    def _find_difference_local(self, comparator):
        differences = {}

        s3_keys = self.s3_wrapper.list_files(self.s3_prefix)

        files = get_file_tree(self.filepath)

        # +1 to remove filepath separator
        local_filepath_length = len(self.filepath) + 1

        for filepath in files:
            path = join(self.s3_prefix, filepath[local_filepath_length:])
            path = path.replace('\\', '/')
            if path in s3_keys:
                modified = comparator(FileSynchronizer.local_file_stats(filepath),
                                      FileSynchronizer.s3_file_stats(s3_keys[path]))

                if modified:
                    differences[filepath] = FileDifference(filepath, path, FileDifference.State.MODIFIED)

                del s3_keys[path]
            else:
                differences[filepath] = FileDifference(filepath, path, FileDifference.State.NEW)

        s3_prefix_length = len(self.s3_prefix)
        for remote_path in s3_keys.keys():
            local_path = join(self.filepath, remote_path[s3_prefix_length:])
            differences[local_path] = FileDifference(local_path, remote_path, FileDifference.State.DELETED)

        to_remove = []
        for file_path, value in differences.items():
            if value.state in self.ignore_file_states:
                to_remove.append(file_path)

        for file_path in to_remove:
            del differences[file_path]

        return differences

    def _find_difference_remote(self, comparator):
        differences = {}

        file_path = self.s3_prefix
        if self.s3_full_path is not None:
            file_path = self.s3_full_path

        s3_keys = self.s3_wrapper.list_files(file_path)
        files = get_file_tree(self.filepath)

        s3_prefix_length = len(self.s3_prefix)

        for key, value in s3_keys.items():
            path = join(self.filepath, key[s3_prefix_length:])

            if sys.platform == "win32":
                path = path.replace("/", "\\")

            if path in files:
                modified = comparator(FileSynchronizer.local_file_stats(path),
                                      FileSynchronizer.s3_file_stats(value))

                if modified:
                    differences[path] = FileDifference(path, key, FileDifference.State.MODIFIED)

                files.remove(path)
            else:
                differences[path] = FileDifference(path, key, FileDifference.State.NEW)

        for local_path in files:
            remote_path = join(self.s3_prefix, local_path[(len(self.filepath) + 1):])
            differences[local_path] = FileDifference(local_path, remote_path, FileDifference.State.DELETED)

        to_remove = []
        for file_path, value in differences.items():
            if value.state in self.ignore_file_states:
                to_remove.append(file_path)

        for file_path in to_remove:
            del differences[file_path]

        return differences

    def synchronize(self, file_differences, hooks=None):
        for difference in file_differences:
            self.synchronize_single(difference, hooks)

    def synchronize_single(self, file_difference, hooks=None, max_threads=None):
        """
        Synchronizes the file between local and remote. Uses the master specified in the constructor.

        :param file_difference:
        :type file_difference FileDifference
        :param hooks
        :param max_threads the maximum number of threads to use for the network operation.
        :return: result of the sync
        """

        if file_difference.state in self.ignore_file_states:
            return

        self.call_hooks(hooks, FileEvent(FileEvent.START, file_difference))
        try:
            if self.master == FileSynchronizer.LOCAL:
                result = self._synchronize_local_master(file_difference, max_threads=max_threads)
            else:
                result = self._synchronize_remote_master(file_difference, max_threads=max_threads)

            self.call_hooks(hooks, FileEvent(FileEvent.FINISHED, file_difference, result))
        except BaseException as exception:
            self.call_hooks(hooks, FileEvent(FileEvent.FAILED, file_difference, exception))

    def _synchronize_local_master(self, file_difference, max_threads=None):
        if file_difference.state == FileDifference.State.NEW:
            return self.s3_wrapper.upload_file(file_difference.source_path, file_difference.destination_path,
                                               max_threads=max_threads)
        elif file_difference.state == FileDifference.State.MODIFIED:
            return self.s3_wrapper.upload_file(file_difference.source_path, file_difference.destination_path,
                                               max_threads=max_threads)
        elif file_difference.state == FileDifference.State.DELETED:
            return self.s3_wrapper.delete_file(file_difference.destination_path)
        elif file_difference.state == FileDifference.State.MOVED:
            return self.s3_wrapper.move_file(file_difference.original_destination_path, file_difference.destination_path)

    def _synchronize_remote_master(self, file_difference, max_threads=None):
        if file_difference.state == FileDifference.State.NEW:
            return self.s3_wrapper.download_file(file_difference.source_path, file_difference.destination_path,
                                                 max_threads=max_threads)
        elif file_difference.state == FileDifference.State.MODIFIED:
            os.remove(file_difference.source_path)
            return self.s3_wrapper.download_file(file_difference.source_path, file_difference.destination_path,
                                                 max_threads=max_threads)
        elif file_difference.state == FileDifference.State.DELETED:
            os.remove(file_difference.source_path)
            return True
        elif file_difference.state == FileDifference.State.MOVED:
            raise NotImplementedError('Unable to tell if a file was moved on s3')

    def shutdown(self):
        self.running = False

    def call_hooks(self, hooks, file_event):
        if not self.running:
            return

        if hooks is None:
            return

        for hook in hooks:
            hook(file_event)

    @staticmethod
    def print_status(cwd_length, master=LOCAL):
        """Returns a function that takes in a file_difference and prints the action on the input file
        Uses the current working directory length to print relative file names.
        """

        def print_status_internal_local(file_event):
            """
             :param file_event:
             :type file_event FileEvent
             :return:
             """
            if file_event.state == FileEvent.START:
                return

            file_difference = file_event.file_difference

            local_path = file_difference.source_path[cwd_length:]

            if file_difference.state == FileDifference.State.MOVED:
                local_source_path = file_difference.original_source_path[cwd_length:]

                if file_event.state == FileEvent.FINISHED:
                    print('move {} to {}'.format(local_source_path, local_path))
                elif file_event.state == FileEvent.FAILED:
                    print('(FAILED) move {} to {}'.format(local_source_path, local_path))
                else:
                    print('move {} to {}'.format(local_source_path, local_path))

                return

            action = 'upload'

            if file_difference.state == FileDifference.State.DELETED:
                action = 'delete'

            if file_difference.state == FileDifference.State.MODIFIED:
                action = 'upload modified'

            if file_event.state == FileEvent.FAILED:
                action = '(FAILED) {} - {}'.format(action, file_event.result)

            print('{}: {}'.format(action, local_path))

        def print_status_internal_remote(file_event):
            """

            :param file_event:
            :type file_event FileEvent
            :return:
            """
            if file_event.state == FileEvent.START:
                return

            file_difference = file_event.file_difference

            # Pick out the version of the remote folder.
            # Want to display account_uid / datasets / dataset_uid / path / to / file.txt
            local_path = file_difference.original_destination_path
            parts = local_path.split('/', 4)
            del parts[3]
            local_path = "/".join(parts)

            if file_difference.state == FileDifference.State.MOVED:
                local_source_path = file_difference.original_source_path[cwd_length:]
                print('move {} to {}'.format(local_source_path, local_path))
                return

            action = 'download'

            if file_difference.state == FileDifference.State.DELETED:
                action = 'delete'

            if file_difference.state == FileDifference.State.MODIFIED:
                action = 'download modified'

            if file_event.state == FileEvent.FAILED:
                action = '(FAILED) {} - {}'.format(action, file_event.result)

            print('{}: {}'.format(action, '/' + local_path))

        if master == FileSynchronizer.LOCAL:
            return print_status_internal_local

        return print_status_internal_remote


class ThreadedFileSynchronizer:
    def __init__(self, synchronizer, hooks=None):
        self.synchronizer = synchronizer
        self.executor = concurrent.futures.ThreadPoolExecutor()
        self.backlog = {}
        self.active_items = {}
        self.hooks = hooks
        self.lock = Lock()
        self.running = True
        self.max_threads = (computer_cpu_count() or 1) * 10

    def find_difference(self, comparator=None):
        return self.synchronizer.find_difference(comparator)

    def synchronize(self, file_differences):
        for difference in file_differences:
            self.synchronize_single(difference)

    def synchronize_single(self, file_difference):
        """
        :param file_difference:
        :type file_difference FileDifference
        :return:
        """
        with self.lock:
            # Ensure operations on the same source file are done in order they come in
            for current_task in self.active_items.values():
                if current_task.original_source_path == file_difference.original_source_path:
                    self._add_to_backlog(file_difference)
                    return

        self._execute_item(file_difference)

    def _execute_item(self, file_difference):
        if not self.running:
            return

        task = self.executor.submit(self.synchronizer.synchronize_single, file_difference, self.hooks, self.max_threads)

        with self.lock:
            self.active_items[task] = file_difference

        task.add_done_callback(self._on_task_completed)

    def shutdown(self, wait=True):
        if not wait:
            self.synchronizer.shutdown()

        self.running = False
        self.backlog = {}
        self.executor.shutdown(wait)

    def _on_task_completed(self, arg):
        with self.lock:
            file_difference = self.active_items[arg]
            del self.active_items[arg]

        backlog = self._get_backlog(file_difference)
        if backlog is not None:
            self._execute_item(backlog)

    def _add_to_backlog(self, file_difference):
        key = file_difference.original_source_path
        if key in self.backlog:
            self.backlog[key].append(file_difference)
            return

        self.backlog[key] = [file_difference]

    def _get_backlog(self, file_difference):
        key = file_difference.original_source_path
        if key not in self.backlog:
            return None

        todo = self.backlog[key]
        if len(todo) == 0:
            return None

        return todo.pop()



from onepanel.models import VolumeType
from onepanel.commands.volume_types import VolumeTypeViewController

class VolumeTypes():
    def __init__(self, conn):
        self.volume_type_view_controller = VolumeTypeViewController(conn)
    
    def list(self):
        vc = self.volume_type_view_controller
        items = vc.list(params='/volume_types')
        if items == None:
            print('No machine types found.')
            return

        vc.print_items(items, fields=['uid','name'], field_names=['UID','SPECS'])
        return [VolumeType.from_json(item) for item in items]


""" Command line interface for the OnePanel Machine Learning platform

Entry point for command line interface.
"""

import click

from onepanel.commands.common import clone, download, push, timer_sync_output
from onepanel.utilities.original_connection import Connection

from onepanel.commands.datasets import datasets
from onepanel.commands.environments import environments
from onepanel.commands.instances import workspaces
from onepanel.commands.jobs import jobs
from onepanel.commands.login import login, login_with_token, logout
from onepanel.commands.machine_types import machine_types
from onepanel.commands.projects import projects
from onepanel.commands.volume_types import volume_types
from onepanel.constants import *

@click.group()
@click.version_option(version=CLI_VERSION, prog_name='Onepanel CLI')
@click.pass_context
def cli(ctx):
    conn = Connection()
    conn.load_credentials()

    ctx.obj['connection'] = conn


cli.add_command(login)
cli.add_command(login_with_token)
cli.add_command(logout)
cli.add_command(clone)
cli.add_command(download)
cli.add_command(push)
cli.add_command(projects)
cli.add_command(datasets)
cli.add_command(jobs)
cli.add_command(machine_types)
cli.add_command(environments)
cli.add_command(volume_types)
cli.add_command(workspaces)
cli.add_command(timer_sync_output)

# System commands 
try:
    from onepanel.commands.system import system
    
    cli.add_command(system)
except ImportError:
    pass


def main():
    return cli(obj={})


if __name__ == '__main__':
    cli(obj={})



import os

from onepanel.utilities.s3.file_sync import FileSynchronizer, ThreadedFileSynchronizer, FileDifference, FileEvent
from onepanel.utilities.s3.authentication import APIProvider, MemoryCachedCredentialsProvider

import onepanel.utilities.s3.wrapper\
    as S3
import onepanel.services as services
from onepanel.utilities.s3.utilities import S3Facade

# TODO document
class CloudStorageUtility:
    cached_credentials_provider = None

    @staticmethod
    def get_facade(account_uid=None, entity_type=None, entity_uid=None, refresh=False):
        s3Wrapper = CloudStorageUtility.get_utility(account_uid, entity_type, entity_uid, refresh)
        return S3Facade(s3Wrapper)

    @staticmethod
    def get_utility(account_uid=None, entity_type=None, entity_uid=None, refresh=False):
        if CloudStorageUtility.cached_credentials_provider is None or refresh:
            connection = services.get_connection()
            endpoint = '{}/accounts/{}/{}/{}/credentials/aws'.format(connection.URL, account_uid,
                                                                        entity_type, entity_uid)
            CloudStorageUtility.cached_credentials_provider = \
                MemoryCachedCredentialsProvider(APIProvider(connection, endpoint))

        return S3.Wrapper(credentials_provider=CloudStorageUtility.cached_credentials_provider)

    @staticmethod
    def get_file_synchronizer(filepath, s3_prefix, s3_wrapper, master=FileSynchronizer.LOCAL, s3_full_path=None, ignore_file_states=None):
        synchronizer = FileSynchronizer(filepath, s3_prefix, s3_wrapper, master, s3_full_path, ignore_file_states)
        return ThreadedFileSynchronizer(synchronizer)

    @staticmethod
    def sync_files(account_uid, entity_type, entity_uid, filepath, s3_prefix, s3_wrapper=None,
                   master=FileSynchronizer.LOCAL, s3_full_path=None, ignore_file_states=None,
                   quiet=False, include=None, prefix=None, refresh=False):
        if s3_wrapper is None:
            s3_wrapper = CloudStorageUtility.get_utility(account_uid, entity_type, entity_uid, refresh)

        synchronizer = CloudStorageUtility.get_file_synchronizer(filepath, s3_prefix, s3_wrapper, master, s3_full_path, ignore_file_states)

        if not quiet:
            filepath_length = len(filepath)
            if master == FileSynchronizer.LOCAL and filepath[-1] != '/':
                filepath_length += 1

            synchronizer.hooks = [FileSynchronizer.print_status(filepath_length, master)]

        differences = synchronizer.find_difference()

        if len(differences) == 0:
            return 0

        if include is not None:
            differences = FileSynchronizer.filter(differences, include, prefix)

        try:
            synchronizer.synchronize(differences.values())
            synchronizer.shutdown()
        except KeyboardInterrupt as e:
            if master == FileSynchronizer.LOCAL:
                print('\nStopping upload. This may take a few seconds')
            else:
                print('\nStopping download. This may take a few seconds')

            synchronizer.shutdown(False)

        return len(differences.values())



import re


def sanitize_dataset_uid(uid = ''):
	suggested = uid.lower()
	suggested = suggested.replace('_', '-')
	suggested = re.sub('[^A-Za-z0-9\-]+', '', suggested)
	# Do not allow uid to start with a non-alphanumeric
	# End must also be alphanumeric
	# todo verify strip operations
	suggested = suggested.strip('-')
	if len(suggested) > 25:
		suggested = suggested[0:25]
	if len(suggested) < 3:
		suggested = suggested.zfill(3)
	return suggested



from onepanel.models.api_json import APIJSON

class ProjectRepository(APIJSON):
    def __init__(self, uid=None, url=None, branch=None):
        self.uid = uid
        self.url = url
        self.branch = branch

    def api_json(self):
        return {
            'uid': self.uid,
            'url': self.url,
            'branch': self.branch
        }

    @classmethod
    def from_json(cls, dct):
        project_repository = cls()
        project_repository.uid = dct['uid']
        project_repository.url = dct['url']
        project_repository.branch = dct['branch']

        return project_repository



from onepanel.models.dataset_mount import DatasetMount
from onepanel.models.environment import Environment
from onepanel.models.instance import Instance
from onepanel.models.job import Job
from onepanel.models.machine_type import MachineType
from onepanel.models.project import Project
from onepanel.models.project_repository import ProjectRepository
from onepanel.models.volume_type import VolumeType

__all__ = ['DatasetMount', 'Environment', 'Instance', 'Job', 'MachineType', 'Project', 'ProjectRepository', 'VolumeType']


from onepanel.utilities.original_connection import Connection


def get_connection():
    """Creates a connection object and loads the default credentials"""
    conn = Connection()
    conn.load_credentials()

    return conn



from onepanel.models import MachineType
from onepanel.commands.machine_types import MachineTypesViewController

class MachineTypes():
    def __init__(self, conn):
        self.machine_type_view_controller = MachineTypesViewController(conn)
    
    def list(self):
        vc = self.machine_type_view_controller
        items = vc.list(params='/machine_types')
        if items == None:
            print('No machine types found.')
            return

        vc.print_items(items, fields=['uid','name'], field_names=['UID','SPECS'])
        return [MachineType.from_json(item) for item in items]


import os
import concurrent.futures
import threading

from onepanel.utilities.s3.wrapper import Wrapper as S3Wrapper
from onepanel.utilities.python_bridge import computer_cpu_count


class ThreadsafeGenerator:
    """Takes an generator and makes it thread-safe by
    wrapping the call to next with a lock."""
    def __init__(self, generator):
        self.generator = generator
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def next(self):
        with self.lock:
            return next(self.generator)


class AtomicInteger:
    def __init__(self, value=0):
        self._value = value
        self._lock = threading.Lock()

    @property
    def value(self):
        with self._lock:
            return self._value

    def increase(self, amount=1):
        with self._lock:
            self._value += amount

    def decrease(self, amount=1):
        self.increase(-amount)


class SourceFuture:
    def __init__(self,  future, source):
        self.future = future
        self.source = source
        self._done_callbacks = []
        self.future.add_done_callback(self.on_done)

    def add_done_callback(self, fn):
        if not self.future.done():
            self._done_callbacks.append(fn)
            return

        fn(self.source, self.future)

    def on_done(self, result):
        for callback in self._done_callbacks:
            callback(self.source, result)


class FileActionResult:
    def __init__(self, key, action):
        self.key = key
        self.action = action

    def __str__(self):
        return '{}: {}'.format(self.action, self.key)


class S3Uploader:
    """
    Issues:
     * User cancels upload (ctrl + c). Doesn't immediately stop.

    """
    def __init__(self, s3_wrapper, workers=None, upload_threads=None):
        """

        :param s3_wrapper:
        :type s3_wrapper S3Wrapper
        :param workers: Number of threads to upload files. If we have 1 thread, we upload 1 file at a time. 5 threads,
                        5 files at a time.

                        Defaults to cpu count * 10

        :param upload_threads: Number of threads per file upload. If we have 5 threads, we can use up to 5 threads to
                               upload a single file.

                               Defaults to cpu count * 10
        """
        self.s3_wrapper = s3_wrapper

        if workers is None:
            workers = (computer_cpu_count() or 1) * 10

        if upload_threads is None:
            upload_threads = (computer_cpu_count() or 1) * 10

        self.upload_threads = upload_threads
        self._active = True
        self.callbacks = []
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=workers)
        self._current_uploads = AtomicInteger()
        self._max_uploads = workers * 2
        self.stop_flag = threading.Event()

        self.file_paths = None
        self.to_directory = None
        self.files_uploaded = 0

    def _format_path(self, path, to_directory):
        """
        Forms to_directory / path.

        Takes care of issues with forward slashes so we have a consistent,
        a/b/c/d/e.

        Note there is no leading slash - per s3 format.

        Also, windows paths with backslashes '\', are converted to forward slashes.

        :param path:
        :param to_directory:
        :return:
        """
        if to_directory[0] == '/':
            to_directory = to_directory[1:]

        if to_directory[-1] != '/':
            to_directory += '/'

        path.replace('\\', '/')

        if path[0] == '/':
            path = path[1:]

        return to_directory + path

    def upload_files(self, file_paths, to_directory='/', *callbacks):
        """

        :param file_paths: a generator of file paths.
        :param to_directory: the s3 path prefix to upload to.
        :param callbacks
        :return:
        """

        if not isinstance(to_directory, str):
            to_directory = str(to_directory)

        self.to_directory = to_directory
        self.file_paths = ThreadsafeGenerator(file_paths)

        if callbacks is not None:
            self.callbacks = callbacks

        self._process_upload()

    def _upload(self, file_path):
        """
        Uploads the file. file_path is uploaded to self.to_directory

        Precondition: self.lock is locked. This is so we can thread-safely increment the _current_uploads
        :param file_path:
        :return:
        """
        self._current_uploads.increase()

        key = self._format_path(file_path, self.to_directory)
        task = self.executor.submit(self.s3_wrapper.upload_file, file_path, key, self.upload_threads)
        source_task = SourceFuture(task, FileActionResult(file_path, 'upload'))

        source_task.add_done_callback(self._done_upload)

        for callback in self.callbacks:
            source_task.add_done_callback(callback)

    def _done_upload(self, source, result):
        self._current_uploads.decrease()
        self.files_uploaded += 1

        self._process_upload()

    def _process_upload(self):
        """
        Submits files to be uploaded. Ensures we do not go over the _max_uploads.

        This request is ignored if the uploader has been shutdown.

        :return:
        """

        if not self._active:
            return

        if self._current_uploads.value >= self._max_uploads:
            return

        while self._current_uploads.value < self._max_uploads:
            try:
                file_path = self.file_paths.next()
            except StopIteration:
                self._active = False
                break
            else:
                self._upload(file_path)

    def shutdown(self, wait=True):
        if not wait:
            self.stop_flag.set()
            self._active = False
            self.executor.shutdown(False)
            return

        while not self.stop_flag.wait(0.1):
            if self._current_uploads.value == 0:
                self.stop_flag.set()

        self.executor.shutdown(False)


class S3CheckUploader(S3Uploader):
    """
    Checks to see if the file is already present remotely.
    If it is, and it is the same, doesn't upload it.
    """
    def __init__(self, s3_wrapper, workers=None, upload_threads=None):
        S3Uploader.__init__(self, s3_wrapper, workers, upload_threads)
        self.existing_files = {}

    def upload_files(self, file_paths, to_directory='/', *callbacks):
        self.existing_files = self.s3_wrapper.list_files(to_directory)

        S3Uploader.upload_files(self, file_paths, to_directory, *callbacks)

    def _upload(self, file_path):
        key = self._format_path(file_path, self.to_directory)

        # Don't upload the file if its the same as the local one.
        # same means: file name and file size are the same.
        if key in self.existing_files:
            remote_size = self.existing_files[key]['Size']
            existing_size = os.path.getsize(file_path)

            if remote_size == existing_size:
                return

        S3Uploader._upload(self, file_path)


class S3Facade:
    def __init__(self, s3_wrapper, max_threads=None):
        """
        :param s3_wrapper:
        :type s3_wrapper S3Wrapper
        :param max_threads the maximum number of threads to allow per operation. Defaults to (cpu count or 1) * 10
        :type max_threads int
        """
        self.s3_wrapper = s3_wrapper

        if max_threads is None:
            max_threads = (computer_cpu_count() or 1) * 10

        self.max_threads = max_threads

    def _format_path(self, path, to_directory):
        """
        Forms to_directory / path.

        Takes care of issues with forward slashes so we have a consistent,
        a/b/c/d/e.

        Note there is no leading slash - per s3 format.

        Also, windows paths with backslashes '\', are converted to forward slashes.

        :param path:
        :param to_directory:
        :return:
        """
        if to_directory[0] == '/':
            to_directory = to_directory[1:]

        if to_directory[-1] != '/':
            to_directory += '/'

        path.replace('\\', '/')

        if path[0] == '/':
            path = path[1:]

        return to_directory + path

    def upload_file(self, file_path, to_directory='/'):
        """

        :param file_path:
        :param to_directory: the s3 path prefix to upload to.
        :return:
        """

        key = self._format_path(file_path, to_directory)
        return self.s3_wrapper.upload_file(file_path, key, max_threads=self.max_threads)

    def upload_files(self, file_paths, to_directory='/', *callbacks):
        """

        :param file_paths:
        :param to_directory: the s3 path prefix to upload to.
        :param callbacks
        :return:
        """

        if not isinstance(to_directory, str):
            to_directory = str(to_directory)

        uploader = S3CheckUploader(self.s3_wrapper)
        uploader.upload_files(file_paths, to_directory, *callbacks)
        return uploader

    def delete_not_in(self, paths, to_directory='/', *callbacks):
        """

        :param paths:
        :param  to_directory: the s3 path prefix to delete from.
        :return:
        """
        remote_files = self.s3_wrapper.list_files(to_directory)
        files_to_delete = remote_files.copy()

        for path in paths:
            formatted_path = self._format_path(path, to_directory)

            if formatted_path in remote_files:
                del files_to_delete[formatted_path]

        for key, value in files_to_delete.items():
            self.s3_wrapper.delete_file(key)

            if callbacks is not None:
                for callback in callbacks:
                    callback(FileActionResult(key, 'deleted'))



from onepanel.models import Environment
from onepanel.commands.environments import EnvironmentViewController

class Environments():
    def __init__(self, conn):
        self.machine_type_view_controller = EnvironmentViewController(conn)
    
    def list(self):
        vc = self.machine_type_view_controller
        items = vc.list(params='/instance_templates')
        if items == None:
            print('No environments found.')
            return

        vc.print_items(items, fields=['uid','name'], field_names=['UID','ENVIRONMENT'])
        return [Environment.from_json(item) for item in items]


import sys
import datetime
import iso8601

if sys.version_info >= (3, 2, 0):
    UTC = datetime.timezone.utc
else:
    # Yoinked from python docs
    ZERO = datetime.timedelta(0)
    class Utc(datetime.tzinfo):
        """UTC Timezone

        """
        def utcoffset(self, dt):
            return ZERO

        def tzname(self, dt):
            return "UTC"

        def dst(self, dt):
            return ZERO

        def __repr__(self):
            return "<Utc>"

    UTC = Utc()


def utcnow():
    return iso8601.parse_date(datetime.datetime.utcnow().isoformat())


def parse_date(date_string):
    if date_string is None:
        return None

    return iso8601.parse_date(date_string)
    
def format_timedelta(value):
    """Takes a timedelta value and converts it to dd:hh:mm:ss or hh:mm:ss if days are 0"""

    if value is None:
        return '00:00:00:00'

    days, remainder = divmod(value.total_seconds(), 86400) #86400 seconds in a day
    hours, remainder = divmod(remainder, 3600) #3600 seconds in an hour
    minutes, seconds = divmod(remainder, 60) #60 seconds in a minute

    if days > 0:
        return '{:02}:{:02}:{:02}:{:02}'.format(int(days), int(hours), int(minutes), int(seconds))

    #three spaces for days and colon
    return '   {:02}:{:02}:{:02}'.format(int(hours), int(minutes), int(seconds))